from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.references import TimeDimensionReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from metricflow_semantics.collection_helpers.syntactic_sugar import mf_first_item
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.simple_metric_input import SimpleMetricInput
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec
from metricflow_semantics.specs.simple_metric_input_spec import SimpleMetricInputSpec


@dataclass(frozen=True)
class SimpleMetricInputSpecProperties:
    """Input dataclass for grouping properties of a sequence of `SimpleMetricInputSpec`s."""

    simple_metric_input_specs: Sequence[SimpleMetricInputSpec]
    semantic_model_name: str
    agg_time_dimension: TimeDimensionReference
    agg_time_dimension_grain: TimeGranularity
    non_additive_dimension_spec: Optional[NonAdditiveDimensionSpec] = None

    @staticmethod
    def create_from_simple_metric_inputs(  # noqa: D102
        simple_metric_inputs: Sequence[SimpleMetricInput],
    ) -> SimpleMetricInputSpecProperties:
        if len(simple_metric_inputs) == 0:
            raise ValueError("No simple-metric inputs provided")

        key_to_inputs: defaultdict[_SimpleMetricInputGroupingKey, list[SimpleMetricInput]] = defaultdict(list)
        for simple_metric_input in simple_metric_inputs:
            key_to_inputs[_SimpleMetricInputGroupingKey.create_from_simple_metric_input(simple_metric_input)].append(
                simple_metric_input
            )
        if len(key_to_inputs) > 1:
            raise ValueError(
                LazyFormat(
                    "The given simple-metric inputs do not have the same grouping key", key_to_inputs=key_to_inputs
                )
            )
        common_key = mf_first_item(key_to_inputs)

        return SimpleMetricInputSpecProperties(
            simple_metric_input_specs=tuple(
                SimpleMetricInputSpec(
                    element_name=simple_metric_input.name,
                )
                for simple_metric_input in simple_metric_inputs
            ),
            semantic_model_name=common_key.model_id.model_name,
            agg_time_dimension=TimeDimensionReference(common_key.agg_time_dimension_name),
            agg_time_dimension_grain=common_key.agg_time_dimension_grain,
            non_additive_dimension_spec=common_key.non_additive_dimension_spec,
        )


@fast_frozen_dataclass()
class _SimpleMetricInputGroupingKey:
    model_id: SemanticModelId
    agg_time_dimension_name: str
    agg_time_dimension_grain: TimeGranularity
    non_additive_dimension_spec: Optional[NonAdditiveDimensionSpec]

    @staticmethod
    def create_from_simple_metric_input(  # noqa: D102
        simple_metric_input: SimpleMetricInput,
    ) -> _SimpleMetricInputGroupingKey:
        return _SimpleMetricInputGroupingKey(
            model_id=simple_metric_input.model_id,
            agg_time_dimension_name=simple_metric_input.agg_time_dimension_name,
            agg_time_dimension_grain=simple_metric_input.agg_time_dimension_grain,
            non_additive_dimension_spec=NonAdditiveDimensionSpec.create_from_simple_metric_input(simple_metric_input),
        )



from __future__ import annotations

import logging

from typing_extensions import override

from metricflow_semantics.experimental.dsi.model_object_lookup import (
    ModelObjectLookup,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import EntityAttributeEdge
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.attribute_nodes import (
    KeyAttributeNode,
)
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    JoinedModelNode,
    LocalModelNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class EntityKeySubgraphGenerator(SemanticSubgraphGenerator):
    """Generator that adds edges for entity-key attributes.

    Each entity defined in a semantic model maps to an entity-key attribute node as the name of the entity can be used
    to query the respective values.
    """

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for lookup in self._manifest_object_lookup.model_object_lookups:
            self._add_edges_for_model(lookup, edge_list)

    def _add_edges_for_model(self, lookup: ModelObjectLookup, edge_list: list[SemanticGraphEdge]) -> None:
        model_id = SemanticModelId.get_instance(model_name=lookup.semantic_model.name)
        semantic_model_node = JoinedModelNode.get_instance(model_id)
        local_semantic_model_node = LocalModelNode.get_instance(model_id)

        key_attribute_nodes = [KeyAttributeNode.get_instance(entity.name) for entity in lookup.semantic_model.entities]
        edge_list.extend(
            [
                EntityAttributeEdge.create(
                    tail_node=semantic_model_node,
                    head_node=key_attribute_node,
                )
                for key_attribute_node in key_attribute_nodes
            ]
        )
        edge_list.extend(
            [
                EntityAttributeEdge.create(
                    tail_node=local_semantic_model_node,
                    head_node=key_attribute_node,
                )
                for key_attribute_node in key_attribute_nodes
            ]
        )



from __future__ import annotations

import logging
from contextlib import contextmanager
from contextvars import ContextVar
from functools import cached_property
from typing import Generator, Mapping, Optional

from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple, MappingItem
from metricflow_semantics.dag.id_prefix import IdPrefix
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.helpers.mapping_helpers import mf_items_to_dict, mf_mapping_to_items
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class SequentialId:
    """Returns a sequentially numbered ID based on a prefix."""

    id_prefix: IdPrefix
    index: int

    @property
    def str_value(self) -> str:  # noqa: D102
        return f"{self.id_prefix.str_value}_{self.index}"

    @override
    def __repr__(self) -> str:
        return self.str_value


@fast_frozen_dataclass()
class _IdGenerationState:
    """Keeps track of the next IDs to return for a given prefix.

    Since this is used in a context variable, it can't be mutable or else mutations will be visible between contexts.
    """

    default_start_value: int
    prefix_to_next_value_items: AnyLengthTuple[MappingItem[IdPrefix, int]]

    @cached_property
    def prefix_to_next_value(self) -> Mapping[IdPrefix, int]:
        return mf_items_to_dict(self.prefix_to_next_value_items)

    @staticmethod
    def create(
        default_start_value: int, prefix_to_next_value: Optional[Mapping[IdPrefix, int]] = None
    ) -> _IdGenerationState:
        return _IdGenerationState(
            default_start_value=default_start_value,
            prefix_to_next_value_items=mf_mapping_to_items(prefix_to_next_value or {}),
        )


class _IdGenerationStateStack:
    """A stack that keeps track of the state of ID generation.

    The stack allows the use of context managers to enter sections where ID generation starts at a configured value.
    When entering the section, a new `_IdGenerationState` is pushed on to the stack, and the state at the top of the
    stack is used to generate IDs. When exiting a section, the state is popped off so that ID generation resumes from
    the previous values.

    The bottom of the stack is the one at the lowest index in `_state_stack_items`, and the top of the stack is the highest.

    The "current state" is the state at the top of the stack.
    """

    _state_stack_items: ContextVar[AnyLengthTuple[_IdGenerationState]] = ContextVar(
        "_IdGenerationStateStack__state_stack_items"
    )

    @classmethod
    def _get_stack_items(cls) -> AnyLengthTuple[_IdGenerationState]:
        if cls._state_stack_items.get(None) is None:
            cls._state_stack_items.set((_IdGenerationState.create(default_start_value=0),))
        return cls._state_stack_items.get()

    @classmethod
    def push_state(cls, id_generation_state: _IdGenerationState) -> None:
        cls._state_stack_items.set(cls._get_stack_items() + (id_generation_state,))

    @classmethod
    def pop_state(cls) -> None:
        initial_items = cls._get_stack_items()
        state_stack_size = len(initial_items)
        if state_stack_size <= 1:
            logger.error(
                LazyFormat(
                    "Attempted to pop the last element in the state stack. Since sequential ID generation may not "
                    "be absolutely critical for resolving queries, logging this as an error but it should be "
                    "investigated.",
                    state_stack_size=state_stack_size,
                ),
                stack_info=True,
            )
            return

        cls._state_stack_items.set(initial_items[:-1])

    @classmethod
    def get_current_state(cls) -> _IdGenerationState:
        return cls._get_stack_items()[-1]

    @classmethod
    def replace_current_state(cls, updated_state: _IdGenerationState) -> None:
        """Remove the current state and replace it the new state."""
        cls._state_stack_items.set(cls._get_stack_items()[:-1] + (updated_state,))


class SequentialIdGenerator:
    """Generates sequential ID values based on a prefix."""

    @classmethod
    def create_next_id(cls, id_prefix: IdPrefix) -> SequentialId:  # noqa: D102
        id_generation_state = _IdGenerationStateStack.get_current_state()
        next_index = id_generation_state.prefix_to_next_value.get(id_prefix) or id_generation_state.default_start_value
        updated_state = _IdGenerationState.create(
            default_start_value=id_generation_state.default_start_value,
            prefix_to_next_value={**id_generation_state.prefix_to_next_value, **{id_prefix: next_index + 1}},
        )
        _IdGenerationStateStack.replace_current_state(updated_state)

        return SequentialId(id_prefix=id_prefix, index=next_index)

    @classmethod
    def reset(cls, default_start_value: int = 0) -> None:
        """Resets the numbering of the generated IDs so that it starts at the given value."""
        _IdGenerationStateStack.replace_current_state(
            _IdGenerationState.create(default_start_value=default_start_value)
        )

    @classmethod
    @contextmanager
    def id_number_space(cls, start_value: int) -> Generator[None, None, None]:
        """Open a context where ID generation starts with the given start value.

        On exit, resume ID numbering from prior to entering the context.
        """
        _IdGenerationStateStack.push_state(_IdGenerationState.create(default_start_value=start_value))
        yield None
        _IdGenerationStateStack.pop_state()



from __future__ import annotations

import textwrap
from collections.abc import Sequence
from typing import Dict, Optional


class InformativeUserError(Exception):
    """Raised for user errors.

    The error is actionable by the user or provides user userful information
    as to why it's unactionable (eg., Feature isn't supported)
    """

    pass


class CustomerFacingSemanticException(InformativeUserError):
    """Class of Exceptions that make it to the customer's eyeballs."""

    pass


class UnableToSatisfyQueryError(CustomerFacingSemanticException):  # noqa: D101
    def __init__(self, error_name: str, context: Optional[Dict[str, str]] = None) -> None:
        """Context will be printed as list of items when this is converted to a string."""
        self.error_name = error_name
        self._context = context

    def __str__(self) -> str:  # noqa: D105
        error_lines = ["Unable To Satisfy Query Error: " + self.error_name]
        if self._context:
            for key, value in self._context.items():
                error_lines.append(f"\n{key}:")
                error_lines.append(f"{textwrap.indent(value, prefix='    ')}")
        return "\n".join(error_lines)


class SemanticException(Exception):  # noqa: D101
    pass


class DuplicateMetricError(SemanticException):  # noqa: D101
    pass


class MetricNotFoundError(SemanticException, KeyError):  # noqa: D101
    pass


class InvalidSemanticModelError(SemanticException):  # noqa: D101
    pass


class ExecutionException(Exception):
    """Raised if there are any errors while executing the execution plan."""

    pass


class UnsupportedEngineFeatureError(InformativeUserError, RuntimeError):
    """Raised when the user attempts to use a feature that isn't supported by the data platform."""


class SqlBindParametersNotSupportedError(Exception):
    """Raised when a SqlClient that does not have support for bind parameters receives a non-empty set of params."""


class UnknownMetricError(InformativeUserError):
    """Raised when user input contains metric names that are not known."""

    def __init__(self, metric_names: Sequence[str]) -> None:  # noqa: D107
        name_count = len(metric_names)
        if name_count == 0:
            raise RuntimeError(f"Can't create an {self.__class__.__name__} without metric names")
        elif name_count == 1:
            super().__init__(f"Unknown metric: {repr(metric_names[0])}")
        else:
            super().__init__(f"Unknown metrics: {list(metric_names)}")


class InvalidQuerySyntax(InformativeUserError):
    """Raised when query syntax is invalid. Primarily used in the where clause."""

    def __init__(self, msg: str) -> None:  # noqa: D107
        super().__init__(msg)


class InvalidQueryException(InformativeUserError):
    """Exception thrown when there is an error with the parameters to a MF query."""

    pass


class RenderSqlTemplateException(InformativeUserError):
    """Exception thrown when there is an error rendering a SQL template."""

    pass


class FeatureNotSupportedError(InformativeUserError):
    """Exception thrown when a feature is not implemented."""

    pass


class SemanticManifestConfigurationError(InformativeUserError):
    """Exception thrown when the semantic manifest is not configured correctly."""

    pass



from __future__ import annotations


class MetricFlowException(Exception):
    """Base class for custom exceptions in MF."""

    pass


class MetricFlowInternalError(MetricFlowException):
    """A non-recoverable error due to an issue within MF and not caused by the user.."""

    pass


class InvalidManifestException(MetricFlowException):
    """Raised when an invalid manifest is detected.

    Generally, a semantic manifest is validated before it is passed into the engine. This is useful to raise in
    sanity checks done outside of validations.
    """

    pass


class GraphvizException(MetricFlowException):
    """Raised when there is an error when calling `graphviz` methods."""



from __future__ import annotations

import logging
from collections.abc import Mapping
from functools import cached_property

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols import SemanticModel
from dbt_semantic_interfaces.type_enums import DimensionType, TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.syntactic_sugar import mf_first_non_none_or_raise
from metricflow_semantics.experimental.dsi.entity_lookup import EntityLookup
from metricflow_semantics.experimental.metricflow_exception import InvalidManifestException
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.attribute_pretty_format import AttributeMapping, AttributePrettyFormattable
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class ModelObjectLookup(AttributePrettyFormattable):
    """Similar to `ManifestObjectLookup` but for objects in a `SemanticModel`."""

    def __init__(self, semantic_model: SemanticModel) -> None:  # noqa: D107
        self._semantic_model = semantic_model

    @cached_property
    def semantic_model(self) -> SemanticModel:  # noqa: D102
        return self._semantic_model

    @cached_property
    def model_id(self) -> SemanticModelId:  # noqa: D102
        return SemanticModelId.get_instance(model_name=self._semantic_model.name)

    @cached_property
    def time_dimension_name_to_grain(self) -> Mapping[str, TimeGranularity]:  # noqa: D102
        time_dimension_name_to_grain: dict[str, TimeGranularity] = {}
        for dimension in self._semantic_model.dimensions:
            # Skip non-time dimensions.
            if dimension.type is DimensionType.TIME:
                pass
            elif dimension.type is DimensionType.CATEGORICAL:
                continue
            else:
                assert_values_exhausted(dimension.type)

            type_params = mf_first_non_none_or_raise(
                dimension.type_params,
                error_supplier=lambda: InvalidManifestException(
                    LazyFormat(
                        "`type_params` should not be `None` for a time dimension.",
                        dimension=dimension,
                        semantic_model=self._semantic_model,
                    )
                ),
            )
            time_dimension_name_to_grain[dimension.name] = type_params.time_granularity
        return time_dimension_name_to_grain

    @cached_property
    def entity_lookup(self) -> EntityLookup:  # noqa: D102
        return EntityLookup(self._semantic_model.entities)

    @cached_property
    @override
    def _attribute_mapping(self) -> AttributeMapping:
        return dict(
            **super()._attribute_mapping,
            **{
                attribute_name: getattr(self, attribute_name)
                for attribute_name in (
                    "model_id",
                    "time_dimension_name_to_grain",
                    "entity_lookup",
                )
            },
        )



from __future__ import annotations

import datetime
import gc
import logging
import math
import time
import timeit
from abc import ABC, abstractmethod
from dataclasses import dataclass
from io import StringIO

from typing_extensions import override

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class BenchmarkFunction(ABC):
    """Closure to encapsulate the setup code and the benchmarked code.

    This closure is helpful for authoring tests as the arguments to the `timeit` calls are strings of code
    snippets. Those strings are annoying to author as the IDE inspections don't work / mypy does not do type checks in
    strings. In addition, all imports must be specifically included in the code snippet.

    Using the closure, there's no need to add additional imports as this can make references to objects in the
    caller's frame.

    This is a WIP - there may be other approaches and there may be implications to measuring code performance. One
    issue that is known is that this is not suitable for checking fast snippets as due to the overhead in the
    method call.

    The `__init__` method of implementing classes should have no parameters.
    """

    @abstractmethod
    def run(self) -> None:
        """This should contain the code to run in the benchmark."""
        raise NotImplementedError


class PerformanceBenchmark:
    """Helper to benchmark execution performance using `timeit`."""

    @staticmethod
    def assert_function_performance(
        left_function_class: type[BenchmarkFunction],
        right_function_class: type[BenchmarkFunction],
        min_performance_factor: float,
    ) -> None:
        """Assert that the right function is `min_performance_factor` times faster than the left function.

        As there is an overhead to calling a function in a class, this is not suitable to fast functions.
        """
        attempt_count = 3
        sleep_time = 1.0
        # 50 microseconds.
        min_function_runtime = 50e-6
        left_timer = timeit.Timer(
            timer=time.thread_time,
            setup="left_function = left_function_class()",
            stmt="left_function.run()",
            globals={"left_function_class": left_function_class},
        )
        right_timer = timeit.Timer(
            timer=time.thread_time,
            setup="right_function = right_function_class()",
            stmt="right_function.run()",
            globals={"right_function_class": right_function_class},
        )

        for i in range(attempt_count):
            left_function_runtime = PerformanceBenchmark._call_with_exception_handling(left_timer)
            right_function_runtime = PerformanceBenchmark._call_with_exception_handling(right_timer)

            if any(runtime < min_function_runtime for runtime in (left_function_runtime, right_function_runtime)):
                raise RuntimeError(
                    LazyFormat(
                        "The runtime of at least one of the provided functions is too short."
                        " This means that the runtime will include the function-call overhead as a significant"
                        " component",
                        left_function_runtime=lambda: datetime.timedelta(seconds=left_function_runtime),
                        right_function_runtime=lambda: datetime.timedelta(seconds=right_function_runtime),
                        min_function_runtime=min_function_runtime,
                    )
                )

            performance_factor = left_function_runtime / right_function_runtime

            logger.debug(
                LazyFormat(
                    "Compared performance of left and right functions.",
                    left_function_runtime=lambda: datetime.timedelta(seconds=left_function_runtime),
                    right_function_runtime=lambda: datetime.timedelta(seconds=right_function_runtime),
                    performance_factor=lambda: f"{performance_factor:.2f}",
                    expected_min_performance_factor=lambda: f"{min_performance_factor:.2f}",
                )
            )

            if performance_factor > min_performance_factor:
                return

            if i == attempt_count - 1:
                raise AssertionError(
                    LazyFormat(
                        "The performance factor is lower than the minimum expected.",
                        expected_min_performance_factor=min_performance_factor,
                        actual_performance_factor=lambda: f"{performance_factor:.2f}",
                        attempt_count=attempt_count,
                    )
                )
            else:
                logger.debug(
                    LazyFormat(
                        "Since performance is lower than expected, sleeping and retrying in case it's a fluke",
                        sleep_time=lambda: f"{sleep_time:.2f}s",
                        attempt_num=i + 1,
                        attempt_count=attempt_count,
                    )
                )
                time.sleep(sleep_time)

    @staticmethod
    def _call_with_exception_handling(timer: timeit.Timer) -> float:
        """Wraps the call to `timeit` to include the stack trace as the one raised by `timeit` does not."""
        gc.collect()
        try:
            return timer.timeit(number=1)
        except Exception as e:
            with StringIO() as fp:
                timer.print_exc(fp)
                # noinspection PyUnresolvedReferences
                raise RuntimeError(
                    LazyFormat("Got an exception while executing timed code.", exception=fp.getvalue())
                ) from e


@dataclass(frozen=True)
class _Point:
    x: int
    y: int

    @staticmethod
    def distance(left: _Point, right: _Point) -> float:
        x_delta = left.x - right.x
        y_delta = left.y - right.y
        return math.sqrt(x_delta * x_delta + y_delta * y_delta)


class _ReferenceFunction(BenchmarkFunction, ABC):
    """Function that runs a bunch of dummy operations to provide a reference for relative performance comparisons."""

    def __init__(self, create_object_count: int, function_call_count: int) -> None:
        self._create_object_count = create_object_count
        self._function_call_count = function_call_count

    @override
    def run(self) -> None:
        points = tuple(_Point(x=i, y=i) for i in range(self._create_object_count))
        for i in range(self._function_call_count):
            _Point.distance(points[i], points[self._create_object_count - i - 1])


class OneSecondFunction(_ReferenceFunction):
    """Function that takes about 1s to run."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(create_object_count=1_000_000, function_call_count=1_000_000)



from __future__ import annotations

import logging
from typing import Optional, Sequence, Union

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MetricReference

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.group_by_item.resolution_dag.input_metric_location import InputMetricDefinitionLocation
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    ComplexMetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.simple_metric_source_node import (
    SimpleMetricGroupByItemSourceNode,
)

logger = logging.getLogger(__name__)


class GroupByItemResolutionDagBuilder:
    """Builds a GroupByItemResolutionDag that can be used to resolve group-by-item specs from spec patterns."""

    def __init__(self, manifest_lookup: SemanticManifestLookup) -> None:  # noqa: D107
        self._manifest_lookup = manifest_lookup

    def _build_dag_component_for_metric(
        self,
        metric_reference: MetricReference,
        metric_input_location: Optional[InputMetricDefinitionLocation],
    ) -> Union[SimpleMetricGroupByItemSourceNode, ComplexMetricGroupByItemResolutionNode]:
        """Builds a DAG component that represents the resolution flow for a metric."""
        metric = self._manifest_lookup.metric_lookup.get_metric(metric_reference)

        metric_inputs = MetricLookup.metric_inputs(metric, include_conversion_metric_input=False)

        if len(metric_inputs) == 0:
            return SimpleMetricGroupByItemSourceNode.create(metric_reference, metric_input_location)

        parent_nodes: list[Union[SimpleMetricGroupByItemSourceNode, ComplexMetricGroupByItemResolutionNode]] = []
        for i, metric_input in enumerate(metric_inputs):
            metric_input_reference = MetricReference(metric_input.name)
            parent_nodes.append(
                self._build_dag_component_for_metric(
                    metric_reference=metric_input_reference,
                    metric_input_location=InputMetricDefinitionLocation(
                        metric_reference,
                        input_metric_list_index=i,
                    ),
                )
            )

        return ComplexMetricGroupByItemResolutionNode.create(
            metric_reference=metric_reference,
            metric_input_location=metric_input_location,
            parent_nodes=tuple(parent_nodes),
        )

    def _build_dag_component_for_query(
        self, metric_references: Sequence[MetricReference], where_filter_intersection: WhereFilterIntersection
    ) -> QueryGroupByItemResolutionNode:
        """Builds a DAG component that represents the resolution flow for a query."""
        if len(metric_references) == 0:
            return QueryGroupByItemResolutionNode.create(
                parent_nodes=(NoMetricsGroupByItemSourceNode.create(),),
                metrics_in_query=metric_references,
                where_filter_intersection=where_filter_intersection,
            )
        return QueryGroupByItemResolutionNode.create(
            parent_nodes=tuple(
                self._build_dag_component_for_metric(
                    metric_reference=metric_reference,
                    metric_input_location=None,
                )
                for metric_reference in metric_references
            ),
            metrics_in_query=metric_references,
            where_filter_intersection=where_filter_intersection,
        )

    def build(
        self, metric_references: Sequence[MetricReference], where_filter_intersection: Optional[WhereFilterIntersection]
    ) -> GroupByItemResolutionDag:
        """Build a resolution DAG for a query.

        Args:
            metric_references: The metrics in the query.
            where_filter_intersection: The filters in the query.

        Returns:
            The associated group-by-item resolution DAG.
        """
        return GroupByItemResolutionDag(
            sink_node=self._build_dag_component_for_query(
                metric_references=metric_references,
                where_filter_intersection=where_filter_intersection
                or PydanticWhereFilterIntersection(where_filters=[]),
            )
        )



from __future__ import annotations

from abc import ABC
from functools import cached_property

from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import ComparisonKey
from metricflow_semantics.experimental.mf_graph.formatting.dot_attributes import (
    DotColor,
    DotNodeAttributeSet,
)
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.node_descriptor import MetricFlowGraphNodeDescriptor
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    GroupByAttributeLabel,
    KeyAttributeLabel,
    TimeClusterLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_constant import ClusterNameFactory
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphNode
from metricflow_semantics.experimental.semantic_graph.sg_node_grouping import SemanticGraphNodeTypedCollection
from metricflow_semantics.experimental.singleton import Singleton
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.time.granularity import ExpandedTimeGranularity


@fast_frozen_dataclass(order=False)
class AttributeNode(SemanticGraphNode, ABC):
    """ABC for attribute nodes. See `SemanticGraph` for additional context on usage."""

    attribute_name: str

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.attribute_name,)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"{self.__class__.__name__}({self.attribute_name})",
            cluster_name=None,
        )

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(AttributeNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(color=DotColor.SALMON_PINK)
        return dot_node

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((GroupByAttributeLabel.get_instance(),))

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=self.attribute_name,
        )


@fast_frozen_dataclass(order=False)
class TimeAttributeNode(AttributeNode, Singleton):
    """An attribute node that corresponds to the different time grains available for querying time dimensions.

    e.g. the graph would contain instances for `day`, `dow`, `month`...
    """

    element_property_additions: FrozenOrderedSet[GroupByItemProperty]

    @classmethod
    def get_instance_for_time_grain(cls, time_grain: TimeGranularity) -> TimeAttributeNode:  # noqa: D102
        return cls._get_instance(
            attribute_name=time_grain.value,
            element_property_additions=FrozenOrderedSet(),
        )

    @classmethod
    def get_instance_for_date_part(cls, date_part: DatePart) -> TimeAttributeNode:  # noqa: D102
        return cls._get_instance(
            attribute_name=StructuredLinkableSpecName.date_part_suffix(date_part),
            element_property_additions=FrozenOrderedSet((GroupByItemProperty.DATE_PART,)),
        )

    @classmethod
    def get_instance_for_expanded_time_grain(  # noqa: D102
        cls,
        expanded_time_grain: ExpandedTimeGranularity,
    ) -> TimeAttributeNode:
        return cls._get_instance(
            attribute_name=expanded_time_grain.name,
            element_property_additions=FrozenOrderedSet((GroupByItemProperty.DERIVED_TIME_GRANULARITY,)),
        )

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"TimeAttribute({self.attribute_name})",
            cluster_name=ClusterNameFactory.TIME,
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return super(TimeAttributeNode, self).labels.union((TimeClusterLabel.get_instance(),))

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=self.attribute_name,
            add_properties=tuple(self.element_property_additions),
            set_element_type=LinkableElementType.TIME_DIMENSION,
        )

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.time_attribute_nodes.add(self)


@fast_frozen_dataclass(order=False)
class KeyAttributeNode(AttributeNode, Singleton):
    """Represents the attribute corresponding to the value of a configured entity.

    In the current MF interface, a "query for an entity" means a query for the column values associated with the
    configured entity in a semantic model. To avoid confusion between entities in the semantic graph and entities
    configured in semantic models, "query for an entity" is phrased as "query for an entity key" in the semantic graph
    module.

    Consider using the phrase `entity value` instead.
    """

    @classmethod
    def get_instance(cls, entity_name: str) -> KeyAttributeNode:  # noqa: D102
        return cls._get_instance(attribute_name=entity_name)

    @property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"KeyAttribute({self.attribute_name})",
            cluster_name=ClusterNameFactory.KEY,
        )

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=self.attribute_name,
            add_properties=(GroupByItemProperty.ENTITY,),
            set_element_type=LinkableElementType.ENTITY,
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return super(KeyAttributeNode, self).labels.union((KeyAttributeLabel.get_instance(),))

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.key_attribute_nodes.add(self)


@fast_frozen_dataclass(order=False)
class CategoricalDimensionAttributeNode(AttributeNode, Singleton):
    """Represents a dimension in a semantic model with `DimensionType.CATEGORICAL`."""

    @classmethod
    def get_instance(cls, dimension_name: str) -> CategoricalDimensionAttributeNode:  # noqa: D102
        return cls._get_instance(attribute_name=dimension_name)

    @property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"Dimension({self.attribute_name})", cluster_name=ClusterNameFactory.DIMENSION
        )

    @property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=self.attribute_name,
            set_element_type=LinkableElementType.DIMENSION,
        )

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.categorical_dimension_attribute_nodes.add(self)



from __future__ import annotations

import logging
from typing import FrozenSet, Optional, Sequence, Set

from metricflow_semantics.dag.mf_dag import DagId
from metricflow_semantics.helpers.string_helpers import mf_indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.time.time_spine_source import TimeSpineSource

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_analyzer import DataflowPlanAnalyzer
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.plan_conversion.convert_to_sql_plan import ConvertToSqlPlanResult
from metricflow.plan_conversion.to_sql_plan.dataflow_to_cte import DataflowNodeToSqlCteVisitor
from metricflow.plan_conversion.to_sql_plan.dataflow_to_subquery import DataflowNodeToSqlSubqueryVisitor
from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.optimizer.optimization_levels import (
    SqlGenerationOptionSet,
    SqlOptimizationLevel,
)
from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlPlanOptimizer
from metricflow.sql.sql_plan import (
    SqlPlan,
    SqlPlanNode,
)
from metricflow.sql.sql_select_node import SqlSelectStatementNode

logger = logging.getLogger(__name__)


class DataflowToSqlPlanConverter:
    """Generates an SQL query plan from a node in the metric dataflow plan."""

    def __init__(
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_manifest_lookup: SemanticManifestLookup,
    ) -> None:
        """Constructor.

        Args:
            column_association_resolver: controls how columns for instances are generated and used between nested
            queries.
            semantic_manifest_lookup: Self-explanatory.
        """
        self._column_association_resolver = column_association_resolver
        self._semantic_manifest_lookup = semantic_manifest_lookup
        self._metric_lookup = semantic_manifest_lookup.metric_lookup
        self._semantic_model_lookup = semantic_manifest_lookup.semantic_model_lookup
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(
            semantic_manifest_lookup.semantic_manifest
        )
        self._custom_granularity_time_spine_sources = TimeSpineSource.build_custom_time_spine_sources(
            tuple(self._time_spine_sources.values())
        )

    @property
    def column_association_resolver(self) -> ColumnAssociationResolver:  # noqa: D102
        return self._column_association_resolver

    def convert_to_sql_plan(
        self,
        sql_engine_type: SqlEngine,
        dataflow_plan_node: DataflowPlanNode,
        optimization_level: SqlOptimizationLevel = SqlOptimizationLevel.default_level(),
        sql_query_plan_id: Optional[DagId] = None,
        spec_output_order: Sequence[InstanceSpec] = (),
    ) -> ConvertToSqlPlanResult:
        """Create an SQL query plan that represents the computation up to the given dataflow plan node."""
        # In case there are bugs that raise exceptions at higher optimization levels, retry generation at a lower
        # optimization level. Generally skip O0 (unless requested) as that level does not include the column pruner.
        # Without that, the generated SQL can be enormous.
        optimization_levels_to_attempt: Sequence[SqlOptimizationLevel] = sorted(
            # Union handles case if O0 was specifically requested.
            set(
                possible_level
                for possible_level in SqlOptimizationLevel
                if SqlOptimizationLevel.O1 <= possible_level <= optimization_level
            ).union({optimization_level}),
            reverse=True,
        )
        retried_at_lower_optimization_level = False
        logger.debug(
            LazyFormat(
                "Attempting to convert to a SQL plan with optimization levels:",
                optimization_levels_to_attempt=optimization_levels_to_attempt,
            )
        )
        for attempted_optimization_level in optimization_levels_to_attempt:
            try:
                # TODO: Make this a more generally accessible attribute instead of checking against the
                # BigQuery-ness of the engine
                use_column_alias_in_group_by = sql_engine_type is SqlEngine.BIGQUERY

                option_set = SqlGenerationOptionSet.options_for_level(
                    attempted_optimization_level, use_column_alias_in_group_by=use_column_alias_in_group_by
                )

                logger.info(
                    LazyFormat(
                        "Using option set for SQL generation:",
                        optimization_level=optimization_level,
                        option_set=option_set,
                    )
                )

                nodes_to_convert_to_cte: FrozenSet[DataflowPlanNode] = frozenset()
                if option_set.allow_cte:
                    nodes_to_convert_to_cte = self._get_nodes_to_convert_to_cte(dataflow_plan_node)

                result = self.convert_using_specifics(
                    dataflow_plan_node=dataflow_plan_node,
                    sql_query_plan_id=sql_query_plan_id,
                    nodes_to_convert_to_cte=nodes_to_convert_to_cte,
                    optimizers=option_set.optimizers,
                    spec_output_order=spec_output_order,
                )

                if retried_at_lower_optimization_level:
                    logger.error(
                        LazyFormat(
                            "Successfully generated the SQL plan using an optimization level lower than the"
                            " requested one. A lower one was used due to an exception using the requested one. Please "
                            "investigate the cause for the exception.",
                            requested_optimization_level=optimization_level,
                            successful_optimization_level=attempted_optimization_level,
                        )
                    )

                return result

            except Exception as e:
                if attempted_optimization_level is optimization_levels_to_attempt[-1]:
                    logger.error(
                        "Exhausted attempts to generate the SQL without exceptions."
                        " Propagating the most recent exception."
                    )
                    raise e
                retried_at_lower_optimization_level = True
                logger.exception(
                    LazyFormat(
                        "Got an exception while generating the SQL plan. This indicates a bug that should be"
                        " investigated, but retrying at a different optimization level to potentially avoid a"
                        " user-facing error.",
                        attempted_optimization_level=optimization_level,
                    )
                )

        raise RuntimeError("Should have returned a result or raised an exception in the loop.")

    def convert_using_specifics(
        self,
        dataflow_plan_node: DataflowPlanNode,
        sql_query_plan_id: Optional[DagId],
        nodes_to_convert_to_cte: FrozenSet[DataflowPlanNode],
        optimizers: Sequence[SqlPlanOptimizer],
        spec_output_order: Sequence[InstanceSpec],
    ) -> ConvertToSqlPlanResult:
        """Helper method to convert using specific options. Main use case are tests."""
        logger.debug(
            LazyFormat("Converting to SQL", nodes_to_convert_to_cte=[node.node_id for node in nodes_to_convert_to_cte])
        )

        if len(nodes_to_convert_to_cte) == 0:
            # Avoid `DataflowNodeToSqlCteVisitor` code path for better isolation during rollout.
            # Later this branch can be removed as `DataflowNodeToSqlCteVisitor` should handle an empty
            # `dataflow_nodes_to_convert_to_cte`.
            to_sql_subquery_visitor = DataflowNodeToSqlSubqueryVisitor(
                column_association_resolver=self.column_association_resolver,
                semantic_manifest_lookup=self._semantic_manifest_lookup,
                spec_output_order=spec_output_order,
            )
            data_set = to_sql_subquery_visitor.get_output_data_set(dataflow_plan_node)
        else:
            to_sql_cte_visitor = DataflowNodeToSqlCteVisitor(
                column_association_resolver=self.column_association_resolver,
                semantic_manifest_lookup=self._semantic_manifest_lookup,
                nodes_to_convert_to_cte=nodes_to_convert_to_cte,
                spec_output_order=spec_output_order,
            )
            data_set = dataflow_plan_node.accept(to_sql_cte_visitor)
            select_statement = data_set.checked_sql_select_node
            data_set = SqlDataSet(
                instance_set=data_set.instance_set,
                sql_select_node=SqlSelectStatementNode.create(
                    description=select_statement.description,
                    select_columns=select_statement.select_columns,
                    from_source=select_statement.from_source,
                    from_source_alias=select_statement.from_source_alias,
                    cte_sources=tuple(to_sql_cte_visitor.generated_cte_nodes()),
                    join_descs=select_statement.join_descs,
                    group_bys=select_statement.group_bys,
                    order_bys=select_statement.order_bys,
                    where=select_statement.where,
                    limit=select_statement.limit,
                    distinct=select_statement.distinct,
                ),
            )

        sql_node: SqlPlanNode = data_set.sql_node

        for optimizer in optimizers:
            logger.debug(LazyFormat(lambda: f"Applying optimizer: {optimizer.__class__.__name__}"))
            sql_node = optimizer.optimize(sql_node)
            logger.debug(
                LazyFormat(
                    lambda: f"After applying optimizer {optimizer.__class__.__name__}, the SQL query plan is:\n"
                    f"{mf_indent(sql_node.structure_text())}"
                )
            )

        return ConvertToSqlPlanResult(
            instance_set=data_set.instance_set,
            sql_plan=SqlPlan(render_node=sql_node, plan_id=sql_query_plan_id),
        )

    def _get_nodes_to_convert_to_cte(
        self,
        dataflow_plan_node: DataflowPlanNode,
    ) -> FrozenSet[DataflowPlanNode]:
        """Handles logic for selecting which nodes to convert to CTEs based on the request."""
        dataflow_plan = dataflow_plan_node.as_plan()
        nodes_to_convert_to_cte: Set[DataflowPlanNode] = set(DataflowPlanAnalyzer.find_common_branches(dataflow_plan))
        # Additional nodes will be added later.

        return frozenset(nodes_to_convert_to_cte)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple

from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.instances import InstanceSet, InstanceSetTransform, SimpleMetricInputInstance
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.simple_metric_input_spec import SimpleMetricInputSpec
from metricflow_semantics.sql.sql_exprs import SqlColumnReference, SqlColumnReferenceExpression, SqlFunctionExpression

from metricflow.dataflow.builder.aggregation_helper import InstanceAliasMapping
from metricflow.plan_conversion.instance_set_transforms.select_columns import CreateSelectColumnsForInstances
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.sql.sql_plan import SqlSelectColumn


@dataclass(frozen=True)
class CreateAggregatedSimpleMetricInputsResult:
    """Result class for `CreateAggregatedSimpleMetricInputsTransform`."""

    # Columns that should be in the `SELECT` clause.
    select_column_set: SelectColumnSet
    # Columns that should be in the `GROUP BY` clause.
    group_by_column_set: SelectColumnSet


class CreateAggregatedSimpleMetricInputsTransform(InstanceSetTransform[CreateAggregatedSimpleMetricInputsResult]):
    """Create select columns of the form "fct_bookings.bookings AS bookings" for all the instances.

    However, for simple-metric-input columns, convert them into expressions like
    "SUM(fct_bookings.bookings) AS bookings" so that the resulting expressions can be used for aggregations.

    Also add an output alias that conforms to the alias
    """

    def __init__(  # noqa: D107
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        manifest_object_lookup: ManifestObjectLookup,
        alias_mapping: InstanceAliasMapping,
    ) -> None:
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._manifest_object_lookup = manifest_object_lookup
        self._alias_mapping = alias_mapping
        self._create_select_column_transform = CreateSelectColumnsForInstances(
            table_alias=table_alias,
            column_resolver=column_resolver,
        )

    def _make_sql_column_expression_for_aggregation(
        self, instances: Tuple[SimpleMetricInputInstance, ...]
    ) -> SelectColumnSet:
        output_columns: List[SqlSelectColumn] = []
        for instance in instances:
            spec = instance.spec
            output_spec = self._alias_mapping.aliased_spec(spec) or spec

            output_columns.append(
                self._make_aggregation_sql_column_expression(instance=instance, output_spec=output_spec)
            )

        return SelectColumnSet.create(simple_metric_input_columns=output_columns)

    def _make_aggregation_sql_column_expression(
        self, instance: SimpleMetricInputInstance, output_spec: SimpleMetricInputSpec
    ) -> SqlSelectColumn:
        """Convert one simple-metric input instance into a SQL column."""
        # Get the column name of the simple-metric input in the table that we're reading from
        column_name_in_table = instance.associated_column.column_name

        # Create an expression that will aggregate the given simple-metric input.
        simple_metric_input = self._manifest_object_lookup.simple_metric_name_to_input[instance.spec.element_name]

        read_expression = SqlColumnReferenceExpression.create(
            SqlColumnReference(self._table_alias, column_name_in_table)
        )

        # Figure out the aggregation function for the simple metric.
        aggregation_expression = SqlFunctionExpression.build_expression_from_aggregation_type(
            aggregation_type=simple_metric_input.agg,
            sql_column_expression=read_expression,
            agg_params=simple_metric_input.agg_params,
        )

        # Get the output column name for the simple-metric input.

        aggregated_column_association = self._column_resolver.resolve_spec(output_spec)
        aggregated_column_name = aggregated_column_association.column_name

        return SqlSelectColumn(
            expr=aggregation_expression,
            column_alias=aggregated_column_name,
        )

    def transform(self, instance_set: InstanceSet) -> CreateAggregatedSimpleMetricInputsResult:  # noqa: D102
        instance_set_without_simple_metric_inputs = instance_set.without_simple_metric_inputs()
        column_set_without_simple_metric_inputs = self._create_select_column_transform.transform(
            instance_set_without_simple_metric_inputs
        ).select_column_set
        group_by_column_set = self._create_select_column_transform.transform(
            instance_set_without_simple_metric_inputs
        ).select_column_set
        simple_metric_input_column_set = self._make_sql_column_expression_for_aggregation(
            instance_set.simple_metric_input_instances
        )

        return CreateAggregatedSimpleMetricInputsResult(
            select_column_set=column_set_without_simple_metric_inputs.merge(simple_metric_input_column_set),
            group_by_column_set=group_by_column_set,
        )



from __future__ import annotations

from typing import Sequence

from metricflow_semantics.model.semantics.linkable_element_set_base import AnnotatedSpec


class RenderedSpecTracker:
    """Mutable object that is used to record the specs that were rendered in a where filter.

    This is useful for constructing a WhereFilterSpec as it includes a list of specs required by the filter.
    """

    def __init__(self) -> None:  # noqa: D107
        self._rendered_specs: list[AnnotatedSpec] = []

    def record_rendered_spec(self, spec: AnnotatedSpec) -> None:
        """Records a spec that was rendered in a where filter and can be retrieved later through rendered_specs()."""
        self._rendered_specs.append(spec)

    @property
    def rendered_specs(self) -> Sequence[AnnotatedSpec]:
        """Returns specs that were recorded by record_rendered_spec()."""
        return self._rendered_specs



from __future__ import annotations

import logging
from typing import Any, Dict, List, Mapping, Optional, Tuple

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.helpers.string_helpers import mf_indent
from metricflow_semantics.mf_logging.format_option import PrettyFormatOption
from metricflow_semantics.mf_logging.pretty_formatter import MetricFlowPrettyFormatter

logger = logging.getLogger(__name__)


def mf_pformat(  # type: ignore
    obj: Any,
    format_option: Optional[PrettyFormatOption] = None,
) -> str:
    """Print objects in a pretty way for logging / test snapshots.

    In Python 3.10, the pretty printer class will support dataclasses, so we can remove this once we're on
    3.10. Also tried the prettyprint package with dataclasses, but that prints full names for the classes
    e.g. a.b.MyClass and it also always added line breaks, even if an object could fit on one line, so
    preferring to not use that for compactness.

    e.g.
        metricflow.specs.DimensionSpec(
            element_name='country',
            entity_links=()
        ),

    Instead, the below will print something like:

        DimensionSpec(element_name='country', entity_links=())

    Also, this simplifies the object representation in some cases (e.g. Enums) and provides options for a more compact
    string. This is an improvement on pformat_big_objects() in dbt-semantic-interfaces to be more compact and easier
    to read.

    Args:
        obj: The object to convert to string.
        format_option: The option to use for formatting.

    Returns:
        A string representation of the object that's useful for logging / debugging.
    """
    # Since this is used in logging calls, wrap with except so that a bug here doesn't result in something breaking.
    try:
        formatter = MetricFlowPrettyFormatter(format_option)
        return formatter.pretty_format(obj)
    except Exception:
        # This automatically includes the call trace.
        logger.exception("Error pretty printing due to an exception - using str() instead.")
        return str(obj)


@fast_frozen_dataclass()
class PrettyFormatDictOption(PrettyFormatOption):
    """Options for `mf_pformat_dict`.

    If `preserve_raw_strings` is set, and a value of the obj_dict is of type str, then use the value itself, not the
    representation of the string. e.g. if value="foo", then "foo" instead of "'foo'". Useful for values that contain
    newlines.

    If `pad_items_with_newlines` is set , each key / value section is padded with newlines.
    """

    preserve_raw_strings: bool = False
    pad_items_with_newlines: bool = False


def mf_pformat_dict(  # type: ignore
    description: Optional[str] = None,
    obj_dict: Optional[Mapping[str, Any]] = None,
    format_option: PrettyFormatDictOption = PrettyFormatDictOption(preserve_raw_strings=True),
) -> str:
    """Prints many objects in an indented form."""
    description_lines: List[str] = [description] if description is not None else []
    obj_dict = obj_dict or {}
    item_sections = []

    str_converted_dict: Dict[str, str] = {}
    for key, value in obj_dict.items():
        if format_option.preserve_raw_strings and isinstance(value, str):
            value_str = value
        else:
            value_str = mf_pformat(obj=value, format_option=format_option)
        str_converted_dict[str(key)] = value_str

        item_section_lines: Tuple[str, ...]
        if "\n" in value_str:
            item_section_lines = (
                f"{key}:",
                mf_indent(
                    value_str,
                    indent_prefix=format_option.indent_prefix,
                ),
            )
        else:
            item_section_lines = (f"{key}: {value_str}",)
        item_section = "\n".join(item_section_lines)

        if description is None:
            item_sections.append(item_section)
        else:
            item_sections.append(mf_indent(item_section))

    result_as_one_line = _as_one_line(
        description=description, str_converted_dict=str_converted_dict, max_line_length=format_option.max_line_length
    )
    if result_as_one_line is not None:
        return result_as_one_line

    if format_option.pad_items_with_newlines:
        return "\n\n".join(description_lines + item_sections)
    else:
        return "\n".join(description_lines + item_sections)


def _as_one_line(
    description: Optional[str], str_converted_dict: Dict[str, str], max_line_length: Optional[int]
) -> Optional[str]:
    """See if the result can be returned in a compact, one-line representation.

    e.g. for:
      mf_pformat_dict("Example output", {"a": 1, "b": 2})

    Compact output:
      Example output (a=1, b=2)

    Normal output:
      Example output
        a: 1
        b: 2
    """
    items = tuple(f"{key_str}={value_str}" for key_str, value_str in str_converted_dict.items())
    value_in_parenthesis = ", ".join(items)
    result = f"{description}" + (f" ({value_in_parenthesis})" if len(value_in_parenthesis) > 0 else "")

    if "\n" in result or (max_line_length is not None and len(result) > max_line_length):
        return None

    return result



from __future__ import annotations

import cProfile
import logging
from contextlib import contextmanager
from pstats import FunctionProfile, Stats, StatsProfile
from typing import Dict, Generic, Iterator, Optional, TypeVar

from dbt_semantic_interfaces.implementations.base import FrozenBaseModel

from metricflow_semantics.test_helpers.performance.report_formatter import (
    SessionReportTextFormatter,
    TableTextFormatter,
)

logger = logging.getLogger(__name__)


TNumber = TypeVar("TNumber", int, float)


class PerformanceMetricComparison(FrozenBaseModel, Generic[TNumber]):
    """Comparison of a performance metric."""

    a: TNumber
    b: TNumber
    abs: TNumber
    pct: float

    @classmethod
    def from_metrics(cls, a: TNumber, b: TNumber) -> PerformanceMetricComparison[TNumber]:
        """Construct a PerformanceMetricComparison from two values."""
        abs = a - b
        pct = (abs / a) if a != 0 else float("inf")
        return PerformanceMetricComparison(
            a=a,
            b=b,
            abs=abs,
            pct=pct,
        )


class FunctionReport(FrozenBaseModel):
    """Performance report for a given function.

    We're using this class since the original FunctionProfile has terrible naming, no
    documentation and weird typing (ncalls is a string that can optionally contain numbers).

    See: https://github.com/python/cpython/blob/732670d93b9b0c0ff8adde07418fd6f8397893ef/Lib/pstats.py#L58
    """

    function_name: str

    # number of non-recursive calls
    base_calls: int
    # total number of times the function was called, including recursion
    total_calls: int

    # time spent on the function body, not counting function calls
    body_time: float
    per_call_body_time: float

    # total time spent on function, including function calls and recursion
    total_time: float
    per_call_total_time: float

    @classmethod
    def from_function_profile(cls, function_name: str, prof: FunctionProfile) -> FunctionReport:
        """Construct a FunctionReport from a FunctionProfile."""
        # Who thought this would ever be a good idea?
        if "/" in prof.ncalls:
            total_calls_str, base_calls_str = prof.ncalls.split("/")
            total_calls = int(total_calls_str)
            base_calls = int(base_calls_str)
        else:
            base_calls = total_calls = int(prof.ncalls)

        return cls(
            function_name=function_name,
            base_calls=base_calls,
            total_calls=total_calls,
            body_time=prof.tottime,
            per_call_body_time=prof.percall_tottime,
            total_time=prof.cumtime,
            per_call_total_time=prof.percall_cumtime,
        )

    def compare(self, other: FunctionReport) -> FunctionReportComparison:
        """Compare self with other."""
        assert self.function_name == other.function_name, "Cannot compare unrelated functions"

        return FunctionReportComparison(
            function_name=self.function_name,
            a=self,
            b=other,
            base_calls=PerformanceMetricComparison[int].from_metrics(self.base_calls, other.base_calls),
            total_calls=PerformanceMetricComparison[int].from_metrics(self.total_calls, other.total_calls),
            body_time=PerformanceMetricComparison[float].from_metrics(self.body_time, other.body_time),
            per_call_body_time=PerformanceMetricComparison[float].from_metrics(
                self.per_call_body_time, other.per_call_body_time
            ),
            total_time=PerformanceMetricComparison[float].from_metrics(self.total_time, other.total_time),
            per_call_total_time=PerformanceMetricComparison[float].from_metrics(
                self.per_call_body_time, other.per_call_body_time
            ),
        )


class FunctionReportComparison(FrozenBaseModel):
    """Comparison of two function reports."""

    function_name: str

    a: FunctionReport
    b: FunctionReport

    base_calls: PerformanceMetricComparison[int]
    total_calls: PerformanceMetricComparison[int]
    body_time: PerformanceMetricComparison[float]
    per_call_body_time: PerformanceMetricComparison[float]
    total_time: PerformanceMetricComparison[float]
    per_call_total_time: PerformanceMetricComparison[float]


class SessionReport(FrozenBaseModel):
    """Contains aggregated runtime statistics about a single performance session."""

    session_id: str

    total_time: float
    functions: Dict[str, FunctionReport]

    @classmethod
    def from_stats(cls, session_id: str, stats: StatsProfile) -> SessionReport:
        """Construct a SessionReport from a StatsProfile instance."""
        functions: Dict[str, FunctionReport] = {}
        for (
            func_name,
            func_profile,
        ) in stats.func_profiles.items():
            full_name = f"{func_profile.file_name}:{func_profile.line_number} :: {func_name}"
            report = FunctionReport.from_function_profile(full_name, func_profile)
            functions[full_name] = report

        return cls(
            session_id=session_id,
            total_time=stats.total_tt,
            functions=functions,
        )

    def compare(self, other: SessionReport) -> SessionReportComparison:
        """Compare this report with other.

        If some function is present in one report but not the other, the entry for its comparison
        will be None.
        """
        assert self.session_id == other.session_id, "Cannot compare unrelated sessions."

        self_functions = set(self.functions.keys())
        other_functions = set(other.functions.keys())
        all_functions = self_functions.union(other_functions)

        return SessionReportComparison(
            session_id=self.session_id,
            a=self,
            b=other,
            total_time=PerformanceMetricComparison[float].from_metrics(self.total_time, other.total_time),
            functions={
                func_name: (
                    self.functions[func_name].compare(other.functions[func_name])
                    if func_name in self.functions and func_name in other.functions
                    else None
                )
                for func_name in all_functions
            },
        )

    def text_format(self, formatter: Optional[SessionReportTextFormatter] = None) -> str:
        """Format this report using the provided formatter."""
        if formatter is None:
            return TableTextFormatter().format_report(self)
        return formatter.format_report(self)


class SessionReportComparison(FrozenBaseModel):
    """A comparison between two session reports."""

    session_id: str

    a: SessionReport
    b: SessionReport

    total_time: PerformanceMetricComparison[float]
    functions: Dict[str, Optional[FunctionReportComparison]]


class SessionReportSet(FrozenBaseModel):
    """A set of session reports."""

    sessions: Dict[str, SessionReport] = {}

    def add_report(self, report: SessionReport) -> None:
        """Add a report and associate it with the session ID."""
        self.sessions[report.session_id] = report

    def compare(self, other: SessionReportSet) -> SessionReportSetComparison:
        """Compare this report set with other."""
        self_sessions = set(self.sessions.keys())
        other_sessions = set(other.sessions.keys())
        all_sessions = self_sessions.union(other_sessions)

        return SessionReportSetComparison(
            sessions={
                session: (
                    self.sessions[session].compare(other.sessions[session])
                    if session in self.sessions and session in other.sessions
                    else None
                )
                for session in all_sessions
            }
        )


class SessionReportSetComparison(FrozenBaseModel):
    """A comparison between two session report sets.

    If a session ID is not present in A or B, the comparison is None
    """

    sessions: Dict[str, Optional[SessionReportComparison]]


class PerformanceTracker:
    """Track performance metrics."""

    def __init__(self) -> None:
        """Initialize the tracker."""
        self._session_id: Optional[str] = None
        self._last_session_report: Optional[SessionReport] = None
        self.report_set = SessionReportSet()

    @contextmanager
    def session(self, session_id: str) -> Iterator[PerformanceTracker]:
        """Create a new measurement session.

        At session start, all state is fresh and it gets cleaned when the session ends.
        """
        if self._session_id:
            raise ValueError("Cannot create nested sessions.")

        self._session_id = session_id

        with cProfile.Profile() as profiler:
            yield self

        raw_stats = Stats(profiler)
        stats = raw_stats.strip_dirs().get_stats_profile()

        report = SessionReport.from_stats(session_id, stats)
        self._last_session_report = report
        self.report_set.add_report(report)

        self._session_id = None

    @property
    def last_session_report(self) -> SessionReport:
        """Get the report for the last session."""
        if not self._last_session_report:
            raise ValueError("Run the profiler before getting a report.")

        return self._last_session_report



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Iterable, Optional

from dbt_semantic_interfaces.references import (
    MetricReference,
)

from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element_set_base import BaseGroupByItemSet

logger = logging.getLogger(__name__)


class GroupByItemSetResolver(ABC):
    """Resolves available group-by items for simple-metric inputs and metrics."""

    @abstractmethod
    def get_set_for_distinct_values_query(
        self,
        set_filter: Optional[GroupByItemSetFilter] = None,
    ) -> BaseGroupByItemSet:
        """Returns queryable items for a distinct group-by-item values query.

        A distinct group-by-item values query does not include any metrics.
        """
        raise NotImplementedError

    @abstractmethod
    def get_common_set(
        self,
        metric_references: Iterable[MetricReference] = (),
        set_filter: Optional[GroupByItemSetFilter] = None,
        joins_disallowed: bool = False,
    ) -> BaseGroupByItemSet:
        """Gets the set of the valid group-by items common to all inputs."""
        raise NotImplementedError



from __future__ import annotations

import itertools
import logging
from collections import defaultdict
from collections.abc import Set
from functools import cached_property
from typing import Mapping, Optional

from typing_extensions import override

from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import MetricFlowPathfinder
from metricflow_semantics.experimental.mf_graph.path_finding.traversal_profile_differ import TraversalProfileDiffer
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import IndexedDunderName
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.edges.edge_labels import (
    DenyEntityKeyQueryResolutionLabel,
    DenyVisibleAttributesLabel,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import MetricNode, MetricTimeNode
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    LocalModelLabel,
    MetricLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraph,
    SemanticGraphEdge,
    SemanticGraphNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_node_grouping import SemanticGraphNodeTypedCollection
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_descriptor import (
    DunderNameDescriptor,
    EntityKeyQueryForGroupByMetric,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import (
    DunderNameTrie,
    MutableDunderNameTrie,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie_resolver import (
    DunderNameTrieResolver,
    TrieResolutionResult,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.entity_key_resolver import (
    EntityKeyTrieResolver,
)
from metricflow_semantics.helpers.performance_helpers import ExecutionTimer
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType

logger = logging.getLogger(__name__)


class GroupByMetricTrieResolver(DunderNameTrieResolver):
    """Resolves the group-by metrics that are available for the given simple-metric inputs / metrics.

    Group-by metrics are handled separately to break out the resolution of all available group-by items into smaller
    pieces.
    """

    def __init__(  # noqa: D107
        self,
        semantic_graph: SemanticGraph,
        path_finder: MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath],
    ) -> None:
        super().__init__(semantic_graph=semantic_graph, path_finder=path_finder)
        self._local_model_nodes = semantic_graph.nodes_with_labels(self._local_model_label)
        self._metric_nodes = semantic_graph.nodes_with_labels(MetricLabel.get_instance())

        self._nodes_in_path_from_metric_nodes_to_local_model_nodes = semantic_graph.nodes_with_labels(
            MetricLabel.get_instance(),
            LocalModelLabel.get_instance(),
        )
        self._metric_time_node = MetricTimeNode.get_instance()

        self._group_by_metric_element_properties = (
            GroupByItemProperty.JOINED,
            GroupByItemProperty.METRIC,
        )

        self._verbose_debug_logs = False

    @override
    def resolve_trie(
        self, source_nodes: OrderedSet[SemanticGraphNode], element_filter: Optional[GroupByItemSetFilter]
    ) -> TrieResolutionResult:
        execution_timer = ExecutionTimer()
        pathfinder_profile_differ = TraversalProfileDiffer(self._path_finder)

        with execution_timer, pathfinder_profile_differ:
            trie_result = self._resolve_trie(source_nodes=source_nodes, element_filter=element_filter)

        return TrieResolutionResult(
            duration=execution_timer.total_duration,
            traversal_profile=pathfinder_profile_differ.profile_delta,
            dunder_name_trie=trie_result,
        )

    @cached_property
    def _metric_name_to_entity_key_trie(self) -> Mapping[str, DunderNameTrie]:
        return _MetricNameToEntityKeyTrieGenerator(
            semantic_graph=self._semantic_graph,
            path_finder=self._path_finder,
            model_node_to_entity_key_trie=self._model_node_to_entity_key_trie,
        ).generate()

    @cached_property
    def _entity_key_name_to_metric_names(self) -> Mapping[str, OrderedSet[str]]:
        entity_key_name_to_metric_names: dict[str, MutableOrderedSet[str]] = defaultdict(MutableOrderedSet)
        for metric_name, entity_key_trie in self._metric_name_to_entity_key_trie.items():
            for indexed_name, _ in entity_key_trie.name_items():
                entity_key_name_to_metric_names[indexed_name[-1]].add(metric_name)

        return entity_key_name_to_metric_names

    def _resolve_trie(
        self,
        source_nodes: OrderedSet[SemanticGraphNode],
        element_filter: Optional[GroupByItemSetFilter],
    ) -> DunderNameTrie:
        trie_to_update = MutableDunderNameTrie()

        # Find all required local-model nodes.
        metric_names_allow_set: Optional[Set[str]] = None

        if element_filter is not None:
            metric_names_allow_set = element_filter.element_name_allowlist
            # Group-by metrics always have these properties, so return an empty set if the filter doesn't allow them.
            if not element_filter.allow(
                element_name=None, element_properties=(GroupByItemProperty.METRIC, GroupByItemProperty.JOINED)
            ):
                logger.debug(
                    LazyFormat(
                        "Skipping resolution due to the element filter",
                        class_name=self.__class__.__name__,
                        element_filter=element_filter,
                    )
                )
                return trie_to_update

        find_descendants_result = self._path_finder.find_descendants(
            graph=self._semantic_graph,
            source_nodes=source_nodes,
            target_nodes=self._local_model_nodes,
            node_allow_set=self._nodes_in_path_from_metric_nodes_to_local_model_nodes,
            deny_labels={self._deny_visible_attributes_label},
        )
        reachable_local_model_nodes = tuple(find_descendants_result.reachable_target_nodes)
        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Found local-model nodes for source nodes.",
                    source_nodes=source_nodes,
                    reachable_local_model_nodes=reachable_local_model_nodes,
                )
            )

        if len(reachable_local_model_nodes) == 0:
            raise MetricFlowInternalError(
                LazyFormat(
                    "No local-model nodes were found to be reachable from the given source nodes.",
                    source_nodes=source_nodes,
                    local_model_nodes=self._local_model_nodes,
                )
            )

        # Figure out which entity-key queries are possible given the local-model nodes.
        entity_key_trie: DunderNameTrie = MutableDunderNameTrie.intersection_merge_common(
            tuple(
                self._model_node_to_entity_key_trie[reachable_local_model_node]
                for reachable_local_model_node in reachable_local_model_nodes
            )
        )
        # `entity_key_trie` contains all entity-keys like `booking__listing__user`. But we only want `booking`,
        # so use 1 as the max length.
        collected_entity_key_names: list[str] = []
        collected_derived_from_model_ids: list[SemanticModelId] = []

        for indexed_dunder_name, descriptor in entity_key_trie.name_items(max_length=1):
            if len(indexed_dunder_name) != 1:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "The indexed dunder name should have only contained 1 element.",
                        indexed_dunder_name=indexed_dunder_name,
                    )
                )
            collected_entity_key_names.append(indexed_dunder_name[0])
            collected_derived_from_model_ids.extend(descriptor.derived_from_model_ids)

        entity_key_names = FrozenOrderedSet(collected_entity_key_names)
        derived_from_model_ids = FrozenOrderedSet(collected_derived_from_model_ids)

        for entity_key_name in entity_key_names:
            metric_names = self._entity_key_name_to_metric_names[entity_key_name]
            if metric_names is None:
                logger.warning(
                    LazyFormat(
                        "Valid group-by metrics for the given entity-key name not found. this may occur"
                        " when a given entity is not accessible from any metric (known case is related to conversion"
                        " simple-metric inputs)",
                        entity_key_name=entity_key_name,
                        known_key_names=list(self._entity_key_name_to_metric_names.keys()),
                    )
                )
            trie_for_entity = trie_to_update.next_name_element_to_trie[entity_key_name]

            items_to_add_to_trie: list[tuple[IndexedDunderName, DunderNameDescriptor]] = []

            for metric_name in metric_names:
                if metric_names_allow_set is not None and metric_name not in metric_names_allow_set:
                    continue
                entity_key_queries: list[EntityKeyQueryForGroupByMetric] = []
                for indexed_dunder_name, descriptor in self._metric_name_to_entity_key_trie[metric_name].name_items():
                    if indexed_dunder_name[-1] != entity_key_name:
                        continue
                    entity_key_queries.append(
                        EntityKeyQueryForGroupByMetric(
                            entity_key_query=indexed_dunder_name,
                            derived_from_model_ids=descriptor.derived_from_model_ids,
                        )
                    )
                descriptor = DunderNameDescriptor(
                    element_type=LinkableElementType.METRIC,
                    time_grain=None,
                    date_part=None,
                    element_properties=self._group_by_metric_element_properties,
                    origin_model_ids=self._virtual_semantic_model_ids,
                    derived_from_model_ids=tuple(derived_from_model_ids),
                    entity_key_queries_for_group_by_metric=tuple(entity_key_queries),
                )
                items_to_add_to_trie.append(((metric_name,), descriptor))
            trie_for_entity.add_name_items(items_to_add_to_trie)

        return trie_to_update

    @cached_property
    def _model_node_to_entity_key_trie(self) -> Mapping[SemanticGraphNode, DunderNameTrie]:
        resolver = EntityKeyTrieResolver(self._semantic_graph)
        resolver_result = resolver.resolve_entity_key_trie_mapping()
        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Resolved entity-key trie",
                    resolver_result=resolver_result,
                )
            )
        return resolver_result


class _MetricNameToEntityKeyTrieGenerator:
    def __init__(
        self,
        semantic_graph: SemanticGraph,
        path_finder: MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath],
        model_node_to_entity_key_trie: Mapping[SemanticGraphNode, DunderNameTrie],
    ) -> None:
        self._current_graph = semantic_graph
        self._path_finder = path_finder
        self._metric_node_to_entity_key_trie: dict[SemanticGraphNode, MutableDunderNameTrie] = {}
        # Use typed collection to ensure `self._metric_nodes` contains typed nodes.
        # This allows the retrieval of the metric name from the node.
        typed_collection = SemanticGraphNodeTypedCollection.create(
            self._current_graph.nodes_with_labels(MetricLabel.get_instance())
        )
        self._simple_metric_nodes = typed_collection.simple_metric_nodes
        self._complex_metric_nodes = typed_collection.complex_metric_nodes
        self._metric_nodes: FrozenOrderedSet[MetricNode] = FrozenOrderedSet(
            itertools.chain(self._simple_metric_nodes, self._complex_metric_nodes)
        )
        self._verbose_debug_logs = False
        self._model_node_to_entity_key_trie = model_node_to_entity_key_trie
        self._local_model_nodes = self._current_graph.nodes_with_labels(LocalModelLabel.get_instance())
        self._allowed_nodes_for_walking_from_metrics_to_models = self._local_model_nodes.union(self._metric_nodes)

    def _get_entity_key_trie_for_metric_node(
        self,
        metric_node: SemanticGraphNode,
        _metric_nodes_in_definition_path: Set[SemanticGraphNode] = frozenset(),
    ) -> MutableDunderNameTrie:
        """Get the entity-key trie for the given metric node.

        Before calling this method, `_metric_node_to_entity_key_trie` should be populated for simple metrics to prevent
        recursive calls to non-metric nodes.
        """
        entity_key_trie = self._metric_node_to_entity_key_trie.get(metric_node)
        if entity_key_trie is not None:
            return entity_key_trie

        if metric_node in _metric_nodes_in_definition_path:
            raise RuntimeError(
                LazyFormat(
                    "Recursive metric definition detected",
                    metric_node=metric_node,
                    metric_nodes_in_definition_path=_metric_nodes_in_definition_path,
                )
            )

        if metric_node not in self._metric_nodes:
            raise RuntimeError(
                LazyFormat(
                    "Traversal reached a non-metric node",
                    current_node=metric_node,
                    metric_nodes=self._metric_nodes,
                    metric_nodes_in_definition_path=_metric_nodes_in_definition_path,
                )
            )

        current_graph = self._current_graph

        # If an edge to an input metric has the `deny_all_label`, it means that entity keys can't be queried for this
        # metric.
        deny_entity_key_label = DenyEntityKeyQueryResolutionLabel.get_instance()
        deny_visible_label = DenyVisibleAttributesLabel.get_instance()

        visible_parent_metric_nodes: MutableOrderedSet[SemanticGraphNode] = MutableOrderedSet()
        # `hidden_parent_metric_nodes` are input metrics that is used to compute this metric, but does the affect the
        # group-by items for this metric. They are still needed to compute the semantic models that this metric is
        # derived from.
        hidden_parent_metric_nodes: MutableOrderedSet[SemanticGraphNode] = MutableOrderedSet()

        for edge_to_successor in current_graph.edges_with_tail_node(metric_node):
            parent_metric_node = edge_to_successor.head_node
            # If any of the input metrics don't allow querying of the entity key, then the entity key can't be queried
            # for this metric.
            if (
                deny_entity_key_label in edge_to_successor.labels
                or deny_entity_key_label in edge_to_successor.head_node.labels
            ):
                result = MutableDunderNameTrie()
                self._metric_node_to_entity_key_trie[metric_node] = result
                return result
            elif (
                deny_visible_label in edge_to_successor.labels
                or deny_visible_label in edge_to_successor.head_node.labels
            ):
                hidden_parent_metric_nodes.add(parent_metric_node)
            else:
                visible_parent_metric_nodes.add(parent_metric_node)

        if len(visible_parent_metric_nodes) == 0:
            logger.warning(
                LazyFormat(
                    "The given metric does not have any visible parent metrics. This is likely a bug and"
                    " should be investigated, but returning an empty result for now.",
                    metric_node=metric_node,
                )
            )
            return MutableDunderNameTrie()

        key_query_sets_to_intersect: list[MutableDunderNameTrie] = []
        for parent_metric_node in visible_parent_metric_nodes:
            if parent_metric_node not in self._metric_node_to_entity_key_trie:
                self._get_entity_key_trie_for_metric_node(
                    metric_node=parent_metric_node,
                    _metric_nodes_in_definition_path=_metric_nodes_in_definition_path | {metric_node},
                )
            key_query_sets_to_intersect.append(self._metric_node_to_entity_key_trie[parent_metric_node])

        dunder_name_trie = MutableDunderNameTrie.intersection_merge_common(key_query_sets_to_intersect)

        # If there are input metrics that are hidden, find the semantic models that they belong to and add them as
        # one of the models that the result is derived from.
        if len(hidden_parent_metric_nodes) > 0:
            find_descendants_result = self._path_finder.find_descendants(
                graph=self._current_graph,
                source_nodes=hidden_parent_metric_nodes,
                target_nodes=self._local_model_nodes,
                node_allow_set=self._allowed_nodes_for_walking_from_metrics_to_models,
            )
            dunder_name_trie.update_derived_from_model_ids(
                tuple(
                    FrozenOrderedSet(
                        node.recipe_step_to_append.add_model_join
                        for node in find_descendants_result.reachable_target_nodes
                        if node.recipe_step_to_append.add_model_join is not None
                    )
                )
            )

        self._metric_node_to_entity_key_trie[metric_node] = dunder_name_trie

        return dunder_name_trie

    def generate(self) -> Mapping[str, MutableDunderNameTrie]:
        """This return a mapping from the metric name, to a trie representing the valid entity-key queries for that metric.

        The descriptors for the trie include context on the semantic models required to query the entity key for the
        given metric.
        """
        current_graph = self._current_graph
        path_finder = self._path_finder
        simple_metric_nodes = self._simple_metric_nodes

        if len(simple_metric_nodes) == 0:
            raise RuntimeError(
                LazyFormat(
                    "Did not find any simple-metric nodes. This indicates an error in graph construction.",
                    simple_metric_nodes=simple_metric_nodes,
                )
            )

        self._metric_node_to_entity_key_trie = {}

        if self._verbose_debug_logs:
            logger.debug(
                LazyFormat(
                    "Finding entity key queries for simple-metric nodes",
                    simple_metric_nodes=simple_metric_nodes,
                    allowed_nodes_for_walking_from_metrics_to_models=self._allowed_nodes_for_walking_from_metrics_to_models,
                )
            )

        deny_labels = {
            # For metrics that require metric time to be in a query, it's not possible to query just the entity keys.
            DenyEntityKeyQueryResolutionLabel.get_instance(),
            # For conversion metrics, the conversion simple-metric input is effectively hidden and is not used to generate the
            # available group-by items for a metric.
            DenyVisibleAttributesLabel.get_instance(),
        }

        # For each base metric, generate the set of possible entity-key queries. These will be used to generate the
        # set of possible entity-key queries for derived metrics.
        for simple_metric_node in simple_metric_nodes:
            simple_metric_node_set = FrozenOrderedSet((simple_metric_node,))

            result = path_finder.find_descendants(
                graph=current_graph,
                source_nodes=simple_metric_node_set,
                target_nodes=self._local_model_nodes,
                node_allow_set=self._allowed_nodes_for_walking_from_metrics_to_models,
                deny_labels=deny_labels,
            )
            visible_source_model_nodes = result.reachable_target_nodes

            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Found model nodes containing the simple-metric inputs for the given metric. Those model nodes"
                        " will be used for generating the available group-by items.",
                        simple_metric_node=simple_metric_node,
                        visible_source_model_nodes=visible_source_model_nodes,
                    )
                )

            if len(visible_source_model_nodes) == 0:
                # Entity-key attributes can't be queried for cumulative metrics (as described by the deny labels),
                # so set an empty result.
                self._metric_node_to_entity_key_trie[simple_metric_node] = MutableDunderNameTrie()
                continue

            entity_key_trie_intersection_list: list[DunderNameTrie] = []
            for source_model_node in visible_source_model_nodes:
                entity_key_trie_intersection_list.append(self._model_node_to_entity_key_trie[source_model_node])

            # Figure out the models that a metric depends on.
            result = path_finder.find_descendants(
                graph=current_graph,
                source_nodes=simple_metric_node_set,
                target_nodes=self._local_model_nodes,
                node_allow_set=self._allowed_nodes_for_walking_from_metrics_to_models,
            )
            source_model_nodes = result.reachable_target_nodes
            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Found model nodes containing the simple-metric inputs for the given metric. Those model nodes"
                        " will be used for generating the available group-by items.",
                        simple_metric_node=simple_metric_node,
                        visible_source_model_nodes=visible_source_model_nodes,
                    )
                )

            common_source_model_ids: MutableOrderedSet[SemanticModelId] = MutableOrderedSet()
            for model_node in source_model_nodes - visible_source_model_nodes:
                model_id = model_node.recipe_step_to_append.add_model_join
                assert model_id is not None, LazyFormat(
                    "A found node is expected to be a model node, but the recipe step does not include the"
                    " associated model ID. The search may be incorrect and found the wrong node.",
                    model_id=model_id,
                    model_node=model_node,
                )
                common_source_model_ids.add(model_id)

            dunder_name_trie = MutableDunderNameTrie.intersection_merge_common(entity_key_trie_intersection_list)

            dunder_name_trie.update_derived_from_model_ids(
                tuple(common_source_model_ids),
            )
            self._metric_node_to_entity_key_trie[simple_metric_node] = dunder_name_trie

        metric_name_to_entity_key_trie: dict[str, MutableDunderNameTrie] = {}

        for metric_node in self._metric_nodes:
            entity_key_trie = self._get_entity_key_trie_for_metric_node(metric_node)
            metric_name_to_entity_key_trie[metric_node.metric_name] = entity_key_trie

        return metric_name_to_entity_key_trie



from __future__ import annotations

import logging
from abc import ABC
from dataclasses import dataclass
from typing import DefaultDict, Generic, Iterable, TypeVar

from typing_extensions import override

from metricflow_semantics.collection_helpers.syntactic_sugar import mf_flatten
from metricflow_semantics.experimental.mf_graph.graph_id import MetricFlowGraphId, SequentialGraphId
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.mf_graph import (
    EdgeT,
    MetricFlowGraph,
    MetricFlowGraphNode,
    NodeT,
)
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet

logger = logging.getLogger(__name__)

MutableGraphT = TypeVar("MutableGraphT", bound="MutableGraph")


@dataclass
class MutableGraph(Generic[NodeT, EdgeT], MetricFlowGraph[NodeT, EdgeT], ABC):
    """Base class for mutable graphs.

    The graph ID is changed to a new value whenever this changes for easier cache management.
    """

    _graph_id: MetricFlowGraphId
    _nodes: MutableOrderedSet[NodeT]
    _edges: MutableOrderedSet[EdgeT]

    _label_to_nodes: DefaultDict[MetricFlowGraphLabel, MutableOrderedSet[NodeT]]
    _tail_node_to_edges: DefaultDict[MetricFlowGraphNode, MutableOrderedSet[EdgeT]]
    _head_node_to_edges: DefaultDict[MetricFlowGraphNode, MutableOrderedSet[EdgeT]]
    _label_to_edges: DefaultDict[MetricFlowGraphLabel, MutableOrderedSet[EdgeT]]

    _node_to_predecessor_nodes: DefaultDict[MetricFlowGraphNode, MutableOrderedSet[NodeT]]
    _node_to_successor_nodes: DefaultDict[MetricFlowGraphNode, MutableOrderedSet[NodeT]]

    def add_node(self, node: NodeT) -> None:  # noqa: D102
        self.add_nodes((node,))

    def add_nodes(self, nodes: Iterable[NodeT]) -> None:  # noqa: D102
        self._nodes.update(nodes)
        for node in nodes:
            for node_label in node.labels:
                self._label_to_nodes[node_label].add(node)
        self._graph_id = SequentialGraphId.create()

    def add_edge(self, edge: EdgeT) -> None:  # noqa: D102
        self.add_edges((edge,))

    def add_edges(self, edges: Iterable[EdgeT]) -> None:  # noqa: D102
        tail_nodes = [edge.tail_node for edge in edges]
        head_nodes = [edge.head_node for edge in edges]

        nodes_to_add: MutableOrderedSet[NodeT] = MutableOrderedSet()
        nodes_to_add.update(tail_nodes, head_nodes)
        nodes_to_add.difference_update(self.nodes)
        self.add_nodes(nodes_to_add)

        for edge in edges:
            tail_node = edge.tail_node
            head_node = edge.head_node

            self._tail_node_to_edges[tail_node].add(edge)
            self._head_node_to_edges[head_node].add(edge)
            self._node_to_successor_nodes[tail_node].add(head_node)
            self._node_to_predecessor_nodes[head_node].add(tail_node)

        self._edges.update(edges)
        self._graph_id = SequentialGraphId.create()

    def update(self, other: MetricFlowGraph[NodeT, EdgeT]) -> None:
        """Add the nodes and edges to this graph."""
        self.add_nodes(other.nodes)
        self.add_edges(other.edges)
        self._graph_id = SequentialGraphId.create()

    @override
    @property
    def nodes(self) -> OrderedSet[NodeT]:  # noqa: D102
        return self._nodes

    @override
    def nodes_with_labels(self, *graph_labels: MetricFlowGraphLabel) -> OrderedSet[NodeT]:
        return FrozenOrderedSet(mf_flatten(self._label_to_nodes[label] for label in graph_labels))

    @override
    @property
    def edges(self) -> OrderedSet[EdgeT]:
        return self._edges

    @override
    def edges_with_tail_node(self, tail_node: MetricFlowGraphNode) -> OrderedSet[EdgeT]:
        return self._tail_node_to_edges[tail_node]

    @override
    def edges_with_head_node(self, head_node: MetricFlowGraphNode) -> OrderedSet[EdgeT]:
        return self._head_node_to_edges[head_node]

    @override
    def edges_with_label(self, label: MetricFlowGraphLabel) -> OrderedSet[EdgeT]:
        return self._label_to_edges[label]

    @override
    def successors(self, node: MetricFlowGraphNode) -> OrderedSet[NodeT]:
        return self._node_to_successor_nodes[node]

    @override
    def predecessors(self, node: MetricFlowGraphNode) -> OrderedSet[NodeT]:
        return self._node_to_predecessor_nodes[node]

    @override
    @property
    def graph_id(self) -> MetricFlowGraphId:
        return self._graph_id



"""Base classes for modeling a graph.

The graph object in `networkx` was evaluated, but it was not found to be well-typed.
"""
from __future__ import annotations

import itertools
import logging
from abc import ABC, abstractmethod
from functools import cached_property
from typing import ClassVar, Generic, Optional, TypeVar

from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.mf_type_aliases import Pair
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import Comparable
from metricflow_semantics.experimental.mf_graph.formatting.dot_attributes import (
    DotEdgeAttributeSet,
    DotGraphAttributeSet,
    DotNodeAttributeSet,
)
from metricflow_semantics.experimental.mf_graph.formatting.graph_formatter import MetricFlowGraphFormatter
from metricflow_semantics.experimental.mf_graph.formatting.pretty_graph_formatter import PrettyFormatGraphFormatter
from metricflow_semantics.experimental.mf_graph.graph_element import (
    MetricFlowGraphElement,
)
from metricflow_semantics.experimental.mf_graph.graph_id import MetricFlowGraphId
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.node_descriptor import MetricFlowGraphNodeDescriptor
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.mf_logging.format_option import PrettyFormatOption
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import (
    MetricFlowPrettyFormatter,
    PrettyFormatContext,
)

logger = logging.getLogger(__name__)

NodeT = TypeVar("NodeT", bound="MetricFlowGraphNode")
EdgeT = TypeVar("EdgeT", bound="MetricFlowGraphEdge")

NodeT_co = TypeVar("NodeT_co", bound="MetricFlowGraphNode", covariant=True)
EdgeT_co = TypeVar("EdgeT_co", bound="MetricFlowGraphEdge", covariant=True)


class MetricFlowGraphNode(MetricFlowGraphElement, MetricFlowPrettyFormattable, Comparable, ABC):
    """Base class for nodes in a directed graph."""

    _DEFAULT_NODE_LABELS: ClassVar[OrderedSet[MetricFlowGraphLabel]] = FrozenOrderedSet()

    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        """Return this as attributes for a DOT node.

        Args:
            include_graphical_attributes: If set, include attributes for a graphical representation.

        Returns: An attribute set that can be used to create the DOT node.
        """
        return DotNodeAttributeSet.create(
            name=self.node_descriptor.node_name,
        )

    @property
    @abstractmethod
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:  # noqa: D102
        raise NotImplementedError

    @property
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        """Return the labels that can be used for lookups to get this node."""
        return MetricFlowGraphNode._DEFAULT_NODE_LABELS

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        formatter = MetricFlowPrettyFormatter(
            format_option=format_context.formatter.format_option.merge(
                PrettyFormatOption(include_object_field_names=False)
            )
        )
        return formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={
                "node": self.node_descriptor.node_name,
            },
        )


@fast_frozen_dataclass(order=False)
class MetricFlowGraphEdge(MetricFlowGraphElement, MetricFlowPrettyFormattable, Comparable, Generic[NodeT_co], ABC):
    """Base class for edges in a directed graph.

    An edge can be visualized as an arrow that points from the tail node to the head node.
    """

    _DEFAULT_EDGE_LABELS: ClassVar[OrderedSet[MetricFlowGraphLabel]] = FrozenOrderedSet()

    tail_node: NodeT_co
    head_node: NodeT_co

    @cached_property
    def node_pair(self) -> Pair[NodeT_co, NodeT_co]:
        """Return a tuple of the head node and the tail node."""
        return (self.tail_node, self.head_node)

    @property
    @abstractmethod
    def inverse(self) -> Self:
        """Return the inverse edge."""
        raise NotImplementedError()

    def as_dot_edge(self, include_graphical_attributes: bool) -> DotEdgeAttributeSet:
        """Return this as attributes for a DOT edge.

        Args:
            include_graphical_attributes: If set, include attributes for a graphical representation.

        Returns: An attribute set that can be used to create the DOT edge.
        """
        return DotEdgeAttributeSet.create(
            tail_name=self.tail_node.node_descriptor.node_name,
            head_name=self.head_node.node_descriptor.node_name,
        )

    @property
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        """Return the labels that can be used for lookups to get this edge."""
        return MetricFlowGraphEdge._DEFAULT_EDGE_LABELS

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        formatter = MetricFlowPrettyFormatter(
            format_option=format_context.formatter.format_option.merge(
                PrettyFormatOption(include_object_field_names=False)
            )
        )
        return formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={
                "edge_str": f"{self.tail_node.node_descriptor.node_name} -> {self.head_node.node_descriptor.node_name}",
            },
        )

    @cached_property
    def labels_for_path_addition(self) -> OrderedSet[MetricFlowGraphLabel]:
        """Return the labels for this edge and the head node.

        This is useful for collecting labels while building a path by adding an edge.
        """
        return self.labels.union(self.head_node.labels)


class MetricFlowGraph(MetricFlowPrettyFormattable, Generic[NodeT_co, EdgeT_co], ABC):
    """Base class for a directed graph."""

    @property
    @abstractmethod
    def nodes(self) -> OrderedSet[NodeT_co]:
        """Return the set of nodes in the graph."""
        raise NotImplementedError()

    @abstractmethod
    def nodes_with_labels(self, *graph_labels: MetricFlowGraphLabel) -> OrderedSet[NodeT_co]:
        """Return nodes in the graph with any one of the given labels."""
        raise NotImplementedError()

    def node_with_label(self, label: MetricFlowGraphLabel) -> NodeT_co:
        """Finds the node with the given label. If not exactly one if found, an error is raised."""
        nodes = self.nodes_with_labels(label)
        matching_node_count = len(nodes)
        if matching_node_count != 1:
            raise KeyError(
                LazyFormat(
                    "Did not find exactly one node with the given label",
                    matching_node_count=matching_node_count,
                    first_10_nodes=list(itertools.islice(nodes, 10)),
                )
            )
        return next(iter(nodes))

    @property
    @abstractmethod
    def edges(self) -> OrderedSet[EdgeT_co]:
        """Return the set of edges in a graph."""
        raise NotImplementedError()

    @abstractmethod
    def edges_with_tail_node(self, tail_node: MetricFlowGraphNode) -> OrderedSet[EdgeT_co]:
        """Returns edges with the given tail node."""
        raise NotImplementedError()

    @abstractmethod
    def edges_with_head_node(self, tail_node: MetricFlowGraphNode) -> OrderedSet[EdgeT_co]:
        """Returns edges with the given head node."""
        raise NotImplementedError()

    @abstractmethod
    def edges_with_label(self, label: MetricFlowGraphLabel) -> OrderedSet[EdgeT_co]:
        """Return the set of edges in a graph that have the given label."""
        raise NotImplementedError()

    def format(self, formatter: MetricFlowGraphFormatter = PrettyFormatGraphFormatter()) -> str:
        """Return a representation of this graph using the given formatter."""
        return formatter.format_graph(self)

    @abstractmethod
    def successors(self, node: MetricFlowGraphNode) -> OrderedSet[NodeT_co]:
        """Returns successors of the given node.

        Raises `UnknownNodeException` if the node does not exist in the graph.
        """
        return FrozenOrderedSet(edge.head_node for edge in self.edges_with_tail_node(node))

    @abstractmethod
    def predecessors(self, node: MetricFlowGraphNode) -> OrderedSet[NodeT_co]:
        """Returns predecessors of the given node.

        Raises `UnknownNodeException` if the node does not exist in the graph.
        """
        return FrozenOrderedSet(edge.tail_node for edge in self.edges_with_head_node(node))

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> str:
        return format_context.formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={"nodes": self.nodes, "edges": self.edges},
        )

    @abstractmethod
    def as_sorted(self) -> Self:
        """Return a copy of this graph with the nodes and edges sorted."""
        raise NotImplementedError

    def _intersect_edges(self, other: MetricFlowGraph[NodeT_co, EdgeT_co]) -> OrderedSet[EdgeT_co]:
        return self.edges.intersection(other.edges)

    @abstractmethod
    def intersection(self, other: Self) -> Self:  # noqa: D102
        raise NotImplementedError()

    @abstractmethod
    def inverse(self) -> Self:  # noqa: D102
        raise NotImplementedError()

    def adjacent_edges(self, selected_nodes: OrderedSet[NodeT_co]) -> OrderedSet[EdgeT_co]:
        """Return the edges that have an end point in the selected nodes."""
        subgraph_edges = MutableOrderedSet[EdgeT_co]()
        for node in selected_nodes:
            for edge in self.edges_with_tail_node(node):
                if edge.head_node in selected_nodes:
                    subgraph_edges.add(edge)
            for edge in self.edges_with_head_node(node):
                if edge.tail_node in selected_nodes:
                    subgraph_edges.add(edge)
        return subgraph_edges

    def as_dot_graph(self, include_graphical_attributes: bool) -> DotGraphAttributeSet:
        """Return the attributes that should be used when creating a representation of this graph in DOT.

        Note that this does not contain the nodes and edges.
        """
        return DotGraphAttributeSet.create(
            name=self.__class__.__name__,
        )

    @property
    @abstractmethod
    def graph_id(self) -> MetricFlowGraphId:
        """Return a graph ID.

        This ID will be used for caching cases.
        """
        raise NotImplementedError()



from __future__ import annotations

import logging
from abc import ABC
from collections.abc import Mapping
from functools import cached_property
from typing import Optional

from typing_extensions import override

from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)

AttributeMapping = Mapping[str, object]


class AttributePrettyFormattable(MetricFlowPrettyFormattable, ABC):
    """Mixin that allows classes to be pretty-printed with a defined set of attributes."""

    @cached_property
    def _attribute_mapping(self) -> AttributeMapping:
        """Return a mapping from the attribute name to the value for pretty printing."""
        return {}

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return format_context.formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping=self._attribute_mapping,
        )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Callable, Dict, FrozenSet, Sequence, Tuple, TypeVar

from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import NodeId
from metricflow_semantics.instances import InstanceSet
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.sql.sql_table import SqlTable
from typing_extensions import override

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_simple_metric_inputs import AggregateSimpleMetricInputsNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.offset_base_grain_by_custom_grain import OffsetBaseGrainByCustomGrainNode
from metricflow.dataflow.nodes.offset_custom_granularity import OffsetCustomGranularityNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.plan_conversion.instance_set_transforms.select_columns import CreateSelectColumnsForInstances
from metricflow.plan_conversion.to_sql_plan.dataflow_to_subquery import DataflowNodeToSqlSubqueryVisitor
from metricflow.sql.sql_cte_node import SqlCteNode
from metricflow.sql.sql_plan import SqlSelectColumn
from metricflow.sql.sql_select_node import SqlSelectStatementNode
from metricflow.sql.sql_table_node import SqlTableNode

logger = logging.getLogger(__name__)


DataflowNodeT = TypeVar("DataflowNodeT", bound=DataflowPlanNode)


@dataclass(frozen=True)
class CteGenerationResult:
    """This stores parameters for creating a dataset from a CTE."""

    source_dataflow_plan_node_id: NodeId
    cte_node: SqlCteNode
    select_columns: Tuple[SqlSelectColumn, ...]
    instance_set: InstanceSet

    def get_sql_data_set(self) -> SqlDataSet:
        """Return a dataset that represents reading from the CTE."""
        cte_alias = self.cte_node.cte_alias
        return SqlDataSet(
            instance_set=self.instance_set,
            sql_select_node=SqlSelectStatementNode.create(
                description=f"Read From CTE For node_id={self.source_dataflow_plan_node_id}",
                select_columns=self.select_columns,
                from_source=SqlTableNode.create(SqlTable(schema_name=None, table_name=cte_alias)),
                from_source_alias=cte_alias,
            ),
        )


class DataflowNodeToSqlCteVisitor(DataflowNodeToSqlSubqueryVisitor):
    """Similar to `DataflowNodeToSqlSubqueryVisitor`, except that this converts specific nodes to CTEs.

    This is implemented as a subclass of `DataflowNodeToSqlSubqueryVisitor` so that by default, it has the same behavior
    but in cases where there are nodes that should be converted to CTEs, alternate methods can be used.

    The generated CTE nodes are collected instead of getting incorporated into the associated SQL query plan generated
    at each node so that the CTE nodes can be included at the top-level SELECT statement.

    # TODO: Move these visitors to separate files at the end of the stack.
    """

    def __init__(  # noqa: D107
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_manifest_lookup: SemanticManifestLookup,
        nodes_to_convert_to_cte: FrozenSet[DataflowPlanNode],
        spec_output_order: Sequence[InstanceSpec],
    ) -> None:
        super().__init__(
            column_association_resolver=column_association_resolver,
            semantic_manifest_lookup=semantic_manifest_lookup,
            spec_output_order=spec_output_order,
        )
        self._nodes_to_convert_to_cte = nodes_to_convert_to_cte

        # If a given node is supposed to use a CTE, map the node to the result.
        self._node_to_cte_generation_result: Dict[DataflowPlanNode, CteGenerationResult] = {}

    def generated_cte_nodes(self) -> Sequence[SqlCteNode]:
        """Returns the CTE nodes that have been generated while traversing the dataflow plan."""
        return tuple(result.cte_node for result in self._node_to_cte_generation_result.values())

    def _default_handler(
        self,
        node: DataflowNodeT,
        node_to_select_subquery_function: Callable[[DataflowNodeT], SqlDataSet],
        use_spec_output_order: bool,
    ) -> SqlDataSet:
        """Default handler that is called for each node as the dataflow plan is traversed.

        Args:
            node: The current node in traversal.
            node_to_select_subquery_function: A function that converts the given node to a `SqlDataSet` where the
            SELECT statement source is a subquery. This should be a method in `DataflowNodeToSqlSubqueryVisitor` as this
            was the default behavior before CTEs were supported.

        Returns: The `SqlDataSet` that produces the data for the given node.
        """
        # For the given node, if a CTE was generated for it, return the associated data set.
        cte_generation_result = self._node_to_cte_generation_result.get(node)
        if cte_generation_result is not None:
            logger.debug(LazyFormat("Handling node via existing CTE", node=node))
            return cte_generation_result.get_sql_data_set()

        # If the given node is supposed to use a CTE, generate one for it. Otherwise, use the default subquery as the
        # source for the SELECT.
        select_from_subquery_dataset = node_to_select_subquery_function(node)
        if node not in self._nodes_to_convert_to_cte:
            logger.debug(LazyFormat("Handling node via subquery", node=node))
            return select_from_subquery_dataset
        logger.debug(LazyFormat("Handling node via new CTE", node=node))

        cte_alias = f"{node.node_id.id_str}_{StaticIdPrefix.CTE.value}"

        if cte_alias in set(node.cte_alias for node in self.generated_cte_nodes()):
            raise ValueError(
                f"{cte_alias=} is a duplicate of one that already exists. "
                f"This implies a bug that is generating a CTE for the same dataflow plan node multiple times."
            )

        cte_source = SqlCteNode.create(
            select_statement=select_from_subquery_dataset.sql_node,
            cte_alias=cte_alias,
        )

        cte_generation_result = CteGenerationResult(
            source_dataflow_plan_node_id=node.node_id,
            cte_node=cte_source,
            select_columns=CreateSelectColumnsForInstances(
                table_alias=cte_alias,
                column_resolver=self._column_association_resolver,
            )
            .transform(select_from_subquery_dataset.instance_set)
            .get_columns(self._spec_output_order if use_spec_output_order else ()),
            instance_set=select_from_subquery_dataset.instance_set,
        )
        self._node_to_cte_generation_result[node] = cte_generation_result
        return cte_generation_result.get_sql_data_set()

    @override
    def visit_source_node(self, node: ReadSqlSourceNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_source_node, use_spec_output_order=False
        )

    @override
    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_join_on_entities_node, use_spec_output_order=False
        )

    @override
    def visit_aggregate_simple_metric_inputs_node(self, node: AggregateSimpleMetricInputsNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_aggregate_simple_metric_inputs_node,
            use_spec_output_order=False,
        )

    @override
    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_compute_metrics_node, use_spec_output_order=False
        )

    @override
    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_window_reaggregation_node,
            use_spec_output_order=False,
        )

    @override
    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_order_by_limit_node, use_spec_output_order=False
        )

    @override
    def visit_where_constraint_node(self, node: WhereConstraintNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_where_constraint_node, use_spec_output_order=False
        )

    @override
    def visit_write_to_result_data_table_node(self, node: WriteToResultDataTableNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_write_to_result_data_table_node,
            use_spec_output_order=True,
        )

    @override
    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_write_to_result_table_node,
            use_spec_output_order=True,
        )

    @override
    def visit_filter_elements_node(self, node: FilterElementsNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_filter_elements_node, use_spec_output_order=False
        )

    @override
    def visit_combine_aggregated_outputs_node(self, node: CombineAggregatedOutputsNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_combine_aggregated_outputs_node,
            use_spec_output_order=False,
        )

    @override
    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_constrain_time_range_node,
            use_spec_output_order=False,
        )

    @override
    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_join_over_time_range_node,
            use_spec_output_order=False,
        )

    @override
    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_semi_additive_join_node,
            use_spec_output_order=False,
        )

    @override
    def visit_metric_time_dimension_transform_node(self, node: MetricTimeDimensionTransformNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_metric_time_dimension_transform_node,
            use_spec_output_order=False,
        )

    @override
    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_join_to_time_spine_node,
            use_spec_output_order=False,
        )

    @override
    def visit_min_max_node(self, node: MinMaxNode) -> SqlDataSet:
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_min_max_node, use_spec_output_order=False
        )

    @override
    def visit_add_generated_uuid_column_node(self, node: AddGeneratedUuidColumnNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_add_generated_uuid_column_node,
            use_spec_output_order=False,
        )

    @override
    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_join_conversion_events_node,
            use_spec_output_order=False,
        )

    @override
    def visit_join_to_custom_granularity_node(self, node: JoinToCustomGranularityNode) -> SqlDataSet:
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_join_to_custom_granularity_node,
            use_spec_output_order=False,
        )

    @override
    def visit_alias_specs_node(self, node: AliasSpecsNode) -> SqlDataSet:  # noqa: D102
        return self._default_handler(
            node=node, node_to_select_subquery_function=super().visit_alias_specs_node, use_spec_output_order=False
        )

    @override
    def visit_offset_base_grain_by_custom_grain_node(
        self, node: OffsetBaseGrainByCustomGrainNode
    ) -> SqlDataSet:  # noqa: D102
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_offset_base_grain_by_custom_grain_node,
            use_spec_output_order=False,
        )

    @override
    def visit_offset_custom_granularity_node(self, node: OffsetCustomGranularityNode) -> SqlDataSet:  # noqa: D102
        return self._default_handler(
            node=node,
            node_to_select_subquery_function=super().visit_offset_custom_granularity_node,
            use_spec_output_order=False,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.implementations.metric import (
    PydanticMetric,
    PydanticMetricAggregationParams,
    PydanticMetricInput,
    PydanticMetricTypeParams,
)
from dbt_semantic_interfaces.type_enums import AggregationType, MetricType

from metricflow_semantics.test_helpers.synthetic_manifest.simple_metric_semantic_model_generator import (
    SimpleMetricSemanticModelGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


@dataclass(frozen=True)
class MetricIndex:
    """Index for a generated metric in the semantic manifest.

    Since metrics can be defined through other metrics, the `depth_index` describes the number of hops between a metric
    and the simple metrics that it is based on when looking at the definition tree of a metric.

    For example:
    * `depth_index=0` describes a simple metric that does not depend on any other metrics.
    * `depth_index=1` describes a derived metric that is defined using metrics at `depth_index=0`. i.e. a derived metric
    based on simple metrics.
    * `depth_index=n` describes a derived metric that is defined using metrics at `depth_index=n-1`.

    The `width_index` enumerates the nth metric generated for the given depth (name needs improvement).
    """

    depth_index: int
    width_index: int

    def __post_init__(self) -> None:  # noqa: D105
        if self.depth_index < 0:
            raise ValueError(f"{self.depth_index=} should be >= 0")
        if self.width_index < 0:
            raise ValueError(f"{self.width_index=} should be >=0")


class MetricGenerator:
    """Helps generate metrics for the synthetic manifest."""

    def __init__(  # noqa: D107
        self, parameter_set: SyntheticManifestParameterSet, semantic_model_generator: SimpleMetricSemanticModelGenerator
    ) -> None:
        self._parameter_set = parameter_set
        self._semantic_model_generator = semantic_model_generator

    def generate_metrics(self) -> Sequence[PydanticMetric]:  # noqa: D102
        metrics = []
        for depth_index in range(self._parameter_set.max_metric_depth):
            for width_index in range(self._parameter_set.max_metric_width):
                metrics.append(self._generate_metric(MetricIndex(depth_index=depth_index, width_index=width_index)))

        return metrics

    def get_first_index_at_max_depth(self) -> MetricIndex:
        """For the highest possible metric depth in the semantic manifest, return the index of the first metric."""
        return MetricIndex(
            depth_index=self._parameter_set.max_metric_depth - 1,
            width_index=0,
        )

    def get_next_wrapped_width_index(self, metric_index: MetricIndex) -> MetricIndex:
        """Return the index of the next metric at the same depth level."""
        return MetricIndex(
            depth_index=metric_index.depth_index,
            width_index=(metric_index.width_index + 1) % self._parameter_set.max_metric_width,
        )

    def get_metric_name(self, index: MetricIndex) -> str:  # noqa: D102
        return f"metric_{index.depth_index}_{index.width_index:03}"

    def _metric_indexes_at_depth(self, depth_index: int) -> Sequence[MetricIndex]:
        return tuple(
            MetricIndex(depth_index=depth_index, width_index=width_index)
            for width_index in range(self._parameter_set.max_metric_width)
        )

    def _generate_metric(self, metric_index: MetricIndex) -> PydanticMetric:
        if metric_index.depth_index == 0:
            return PydanticMetric(
                name=self.get_metric_name(metric_index),
                type=MetricType.SIMPLE,
                type_params=PydanticMetricTypeParams(
                    measure=None,
                    numerator=None,
                    denominator=None,
                    expr=None,
                    window=None,
                    grain_to_date=None,
                    metrics=None,
                    conversion_type_params=None,
                    cumulative_type_params=None,
                    metric_aggregation_params=PydanticMetricAggregationParams(
                        agg=AggregationType.SUM,
                        agg_params=None,
                        agg_time_dimension="ds",
                        non_additive_dimension=None,
                        semantic_model=self._semantic_model_generator.get_semantic_model_name(
                            int(metric_index.width_index / self._parameter_set.simple_metrics_per_semantic_model)
                            % self._parameter_set.simple_metric_semantic_model_count
                        ),
                    ),
                ),
                description=None,
                filter=None,
                metadata=None,
                config=None,
            )
        else:
            input_metric_names = tuple(
                self.get_metric_name(lower_depth_metric_index)
                for lower_depth_metric_index in self._metric_indexes_at_depth(metric_index.depth_index - 1)
            )
            return PydanticMetric(
                name=self.get_metric_name(metric_index),
                type=MetricType.DERIVED,
                type_params=PydanticMetricTypeParams(
                    metrics=[
                        PydanticMetricInput(
                            name=input_metric_name, filter=None, alias=None, offset_window=None, offset_to_grain=None
                        )
                        for input_metric_name in input_metric_names
                    ],
                    expr=" + ".join(input_metric_names),
                ),
            )



from __future__ import annotations

import logging
from typing import Final, final

from typing_extensions import override

from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


@final
class ClusterNameFactory:
    """Creates cluster names for the semantic graph.

    * Currently, the cluster name is only used to group related nodes in the semantic graph for visualizations.
    * Clusters / subgraphs could be later used to improve semantic-graph traversal.
    * Additional refinement is pending (e.g. switch to subgraphs, use `ClusterId`).
    """

    TIME: Final[str] = "time"
    KEY: Final[str] = "key"
    CONFIGURED_ENTITY: Final[str] = "configured_entity"
    DIMENSION: Final[str] = "dimension"
    METRIC: Final[str] = "metric"
    TIME_DIMENSION: Final[str] = "time_dimension"

    @override
    def __init__(self) -> None:
        """Can't override `__new__` as that results in a type-check error."""
        raise TypeError(LazyFormat("Class should not be instantiated", cls=self.__class__))

    @staticmethod
    def get_name_for_configured_entity(entity_name: str, model_id: SemanticModelId) -> str:  # noqa: D102
        return f"{model_id}.{entity_name}"

    @staticmethod
    def get_name_for_model(model_id: SemanticModelId) -> str:  # noqa: D102
        """Return the cluster name that should be used to group nodes associated with a specific semantic model."""
        return model_id.model_name



from __future__ import annotations

from typing import Callable


def mf_function_patch_target(method: Callable) -> str:
    """Helper to get the target for a patch.

    Useful so that we don't have strings that reference classes that don't get properly updated during refactoring in
    the IDE.

    Example:
        with patch(get_patch_target(DataflowToSqlPlanConverter.convert_using_specifics)):
            ...

    Annotating with `Callable` as `FunctionType` does not seem to work as expected in `mypy`.
    """
    # noinspection PyUnresolvedReferences
    return f"{method.__module__}.{method.__qualname__}"



"""Nodes used in defining an SQL query plan."""

from __future__ import annotations

import logging
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, Sequence

from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId, DagNode, MetricFlowDag
from metricflow_semantics.sql.sql_exprs import SqlColumnReferenceExpression, SqlExpressionNode
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import Self

if typing.TYPE_CHECKING:
    from metricflow.sql.sql_ctas_node import SqlCreateTableAsNode
    from metricflow.sql.sql_cte_node import SqlCteAliasMapping, SqlCteNode
    from metricflow.sql.sql_select_node import SqlSelectStatementNode
    from metricflow.sql.sql_select_text_node import SqlSelectTextNode
    from metricflow.sql.sql_table_node import SqlTableNode

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=False)
class SqlPlanNode(DagNode["SqlPlanNode"], ABC):
    """A node in the SQL plan model.

    In that model:
    * A source node that takes data into the graph e.g. SQL tables or SQL queries in a literal string format.
    * SELECT queries can be modeled as nodes that transform the data.
    * Statements like ALTER TABLE don't fit well, but they could be modeled as just a single sink node.
    * SQL queries in where conditions could be modeled as another SqlQueryPlan.
    * SqlRenderableNode() indicates nodes where plan generation can begin. Generally, this will be all nodes except
      the SqlTableNode() since my_table.my_column wouldn't be a valid SQL query.

    Is there an existing library that can do this?
    """

    @abstractmethod
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        raise NotImplementedError

    @property
    @abstractmethod
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        """If possible, return this as a select statement node."""
        raise NotImplementedError

    @property
    @abstractmethod
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        """If possible, return this as SQL table node."""
        raise NotImplementedError

    @abstractmethod
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        """Return the SELECT columns that are in this node or the closest `SqlSelectStatementNode` of its ancestors.

        * For a SELECT statement node, this is just the columns in the node.
        * For a node that has a SELECT statement node as its only parent (e.g. CREATE TABLE ... AS SELECT ...), this
          would be the SELECT columns in the parent.
        * If not known (e.g. an arbitrary SQL statement as a string), return None.
        * This is used to figure out which columns are needed at a leaf node of the DAG for column pruning.
        * A SQL table could refer to a CTE, so a mapping from the name of the CTE to the CTE node should be provided to
          get the associated SELECT columns.
        """
        raise NotImplementedError

    @abstractmethod
    def copy(self) -> Self:
        """Return a copy of the branch represented by this node.

        The node fields are copied by reference, similar to shallow copying.
        """
        raise NotImplementedError


class SqlPlanNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of an SQL plan.

    See similar visitor DataflowPlanVisitor.
    """

    @abstractmethod
    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_table_node(self, node: SqlTableNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_query_from_clause_node(self, node: SqlSelectTextNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_cte_node(self, node: SqlCteNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


@dataclass(frozen=True)
class SqlSelectColumn:
    """Represents a column in the select clause of an SQL query."""

    expr: SqlExpressionNode
    # Always require a column alias for simplicity.
    column_alias: str

    @staticmethod
    def from_column_reference(table_alias: str, column_name: str) -> SqlSelectColumn:
        """Create a column that selects a column from a table by name."""
        return SqlSelectColumn(
            expr=SqlColumnReferenceExpression.from_column_reference(column_name=column_name, table_alias=table_alias),
            column_alias=column_name,
        )

    def reference_from(self, source_table_alias: str) -> SqlColumnReferenceExpression:
        """Return a column reference expression for this column with a new table alias.

        Useful when you already have access to the select column from a subquery and want to reference it in an outer query.
        """
        return SqlColumnReferenceExpression.from_column_reference(
            column_name=self.column_alias, table_alias=source_table_alias
        )


class SqlPlan(MetricFlowDag[SqlPlanNode]):
    """Model for an SQL statement as a DAG."""

    def __init__(self, render_node: SqlPlanNode, plan_id: Optional[DagId] = None) -> None:
        """initializer.

        Args:
            render_node: The node from which to start rendering the SQL statement.
            plan_id: If specified, use this sql_query_plan_id instead of a generated one.
        """
        self._render_node = render_node
        super().__init__(
            dag_id=plan_id or DagId.from_id_prefix(StaticIdPrefix.SQL_PLAN_PREFIX),
            sink_nodes=[self._render_node],
        )

    @property
    def render_node(self) -> SqlPlanNode:  # noqa: D102
        return self._render_node



from __future__ import annotations

import threading
from abc import ABC
from typing import ClassVar, Optional, Type, TypeVar

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass

SingletonT = TypeVar("SingletonT", bound="Singleton")


@fast_frozen_dataclass()
class Singleton(ABC):
    """Base class for singleton dataclasses.

    * Useful for objects used as keys.
    * Provides significantly faster equality checks / hashing as `id()` can be used (as by Python's default
      object implementations of `__eq__` and `__hash__`).
    * Slower to create due to lookup overhead (see tests).
    * Clunky to use due to the need to implement `get_instance()` manually, and there doesn't seem to be a good way to
      define an abstract method.
    * Pickling not yet supported.
    * Private initializer is simulated through the `_only_init_via_get_instance` field.
    Implementing classes should define a method similar to:

        @classmethod
        def get_instance(cls, ...) -> Self:
            return cls._get_singleton(...)

    """

    # Consider using a `WeakValueDictionary`.
    _instance_dict: ClassVar[dict] = {}
    _instance_dict_lock: ClassVar[threading.RLock] = threading.RLock()

    # Helps to indicate to users that they shouldn't call `__init__`.
    _only_init_via_get_instance: None

    @classmethod
    def _get_instance(cls: Type[SingletonT], **kwargs) -> SingletonT:  # type: ignore[no-untyped-def]
        # Relies on the order of `kwargs` - https://peps.python.org/pep-0468/
        key = (cls, tuple(kwargs.values()))
        instance_dict = cls._instance_dict
        singleton_instance: Optional[SingletonT] = instance_dict.get(key)
        if singleton_instance is not None:
            return singleton_instance

        with cls._instance_dict_lock:
            # Another thread might have created the singleton between the previous check and getting the lock,
            # so check if one exists again.
            singleton_instance = instance_dict.get(key)
            if singleton_instance is None:
                singleton_instance = cls(_only_init_via_get_instance=None, **kwargs)
                instance_dict[key] = singleton_instance
            return singleton_instance



"""Short functions that act as syntactic sugar / allow for easier auto-completion.

In the process of evaluating whether to keep these.
"""
from __future__ import annotations

import itertools
import logging
from typing import Callable, Iterable, Mapping, Optional

from metricflow_semantics.collection_helpers.mf_type_aliases import KeyT, T, ValueT

logger = logging.getLogger(__name__)


mf_flatten = itertools.chain.from_iterable
mf_group_by = itertools.groupby


def mf_first_non_none(*args: Optional[T]) -> Optional[T]:
    """Return the first non-`None` argument.

    This is helpful because using `a or b` to handle optional items does not work properly for false-like values
    (e.g. 0 or empty string).

    An issue with this method is that all args are evaluated, so it shouldn't be used with expensive expressions
    as arguments. Arguments could be replaced with lambdas.
    """
    for arg in args:
        if arg is not None:
            return arg
    return None


def mf_first_non_none_or_raise(*args: Optional[T], error_supplier: Optional[Callable[[], Exception]] = None) -> T:
    """Similar to `mf_first_non_none` but raises an exception if no values are present."""
    for arg in args:
        if arg is not None:
            return arg

    if error_supplier is not None:
        raise error_supplier()

    raise ValueError("Expected at least one non-`None` argument")


def mf_ensure_mapping(optional_mapping: Optional[Mapping[KeyT, ValueT]]) -> Mapping[KeyT, ValueT]:
    """Returns an empty mapping if `optional_mapping` is `None`.

    Useful for default argument handling.
    """
    return optional_mapping if optional_mapping is not None else {}


def mf_first_item(iterable: Iterable[T], error_supplier: Optional[Callable[[], Exception]] = None) -> T:
    """Return the first item in an iterable."""
    try:
        return next(iter(iterable))
    except StopIteration as e:
        if error_supplier is not None:
            raise error_supplier()

        raise KeyError("Can't return the first item as the iterable has no items") from e



from __future__ import annotations

from typing import Sequence

from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimension
from dbt_semantic_interfaces.implementations.elements.entity import PydanticEntity
from dbt_semantic_interfaces.implementations.node_relation import PydanticNodeRelation
from dbt_semantic_interfaces.implementations.semantic_model import PydanticSemanticModel
from dbt_semantic_interfaces.type_enums import DimensionType, EntityType

from metricflow_semantics.test_helpers.synthetic_manifest.categorical_dimension_generator import (
    CategoricalDimensionGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


class DimensionSemanticModelGenerator:
    """Helps generate a semantic model containing dimensions.

    Each of the generated semantic models contain an entity common to the semantic models containing measures so that
    any measure can be queried by any dimension.
    """

    def __init__(  # noqa: D107
        self,
        parameter_set: SyntheticManifestParameterSet,
        categorical_dimension_generator: CategoricalDimensionGenerator,
    ) -> None:
        self._parameter_set = parameter_set
        self._dimension_generator = categorical_dimension_generator

    def generate_semantic_models(self) -> Sequence[PydanticSemanticModel]:  # noqa: D102
        semantic_models = []
        for semantic_model_index in range(self._parameter_set.dimension_semantic_model_count):
            entities = [
                PydanticEntity(
                    name=self._get_dimension_semantic_model_primary_entity_name(semantic_model_index),
                    type=EntityType.PRIMARY,
                ),
                PydanticEntity(
                    name=self._parameter_set.common_entity_name,
                    type=EntityType.UNIQUE,
                ),
            ]

            dimensions = [
                PydanticDimension(
                    name=self._get_dimension_name(
                        index_in_manifest=semantic_model_index,
                        index_in_model=dimension_index,
                    ),
                    type=DimensionType.CATEGORICAL,
                )
                for dimension_index in range(self._parameter_set.categorical_dimensions_per_semantic_model)
            ]

            semantic_model_name = self._get_dimension_semantic_model_name(semantic_model_index)
            semantic_models.append(
                PydanticSemanticModel(
                    name=semantic_model_name,
                    node_relation=PydanticNodeRelation(
                        schema_name="demo",
                        alias=semantic_model_name,
                    ),
                    entities=entities,
                    dimensions=dimensions,
                )
            )

        return semantic_models

    def _get_dimension_semantic_model_name(self, index_in_manifest: int) -> str:
        return f"dimension_model_{index_in_manifest:03}"

    def _get_dimension_semantic_model_primary_entity_name(self, semantic_model_index: int) -> str:
        return f"{self._get_dimension_semantic_model_name(semantic_model_index)}_primary_entity"

    def _get_dimension_name(self, index_in_manifest: int, index_in_model: int) -> str:
        """Get the name of the dimension given the index.

        Args:
            index_in_manifest: The index of the semantic model in the manifest. e.g. the 2nd semantic model in the
            semantic manifest.
            index_in_model: The index of the dimension in the semantic model. e.g. the 2nd dimension in the semantic
            model.

        Returns:
            The name of the dimension given the index.
        """
        return self._dimension_generator.get_dimension_name(
            index_in_manifest * self._parameter_set.categorical_dimensions_per_semantic_model + index_in_model
        )



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Mapping, Set
from functools import cached_property

from dbt_semantic_interfaces.type_enums import EntityType
from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class EntityJoinType:
    """Describe a type of join between semantic models where entities are of the listed types."""

    left_entity_type: EntityType
    right_entity_type: EntityType


@fast_frozen_dataclass()
class JoinModelOnRightDescriptor:
    """How to join one semantic model onto another, using a specific entity and join type."""

    entity_name: str
    join_type: EntityJoinType


class SemanticModelJoinLookup:
    """Lookup for valid joins between semantic models.

    * This is a simplified / refactored version of the `SemanticModelJoinEvaluator` with no significant logic changes.
    * Migrating `SemanticModelJoinEvaluator` to this one will be done after the semantic graph is fully in place to
      ensure that we have a reference path to compare results in case of differences during rollout.
    """

    _VALID_ENTITY_JOINS: Set[EntityJoinType] = {
        EntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.NATURAL),
        EntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.PRIMARY),
        EntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE),
        EntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.NATURAL),
        EntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.PRIMARY),
        EntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.UNIQUE),
        EntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.NATURAL),
        EntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.PRIMARY),
        EntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.UNIQUE),
        EntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.PRIMARY),
        EntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.UNIQUE),
    }

    _INVALID_ENTITY_JOINS: Set[EntityJoinType] = {
        EntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.FOREIGN),
        EntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.FOREIGN),
        EntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.FOREIGN),
        EntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.FOREIGN),
        # Natural -> Natural joins are not allowed due to hidden fanout or missing value concerns with
        # multiple validity windows in play
        EntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.NATURAL),
    }

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        self._model_id_to_lookup = manifest_object_lookup.model_id_to_lookup
        self._entity_name_to_model_lookups = manifest_object_lookup.entity_name_to_model_lookups
        self._entity_name_to_model_ids = manifest_object_lookup.entity_name_to_model_ids

    @cached_property
    def valid_join_to_entity_types(self) -> OrderedSet[EntityType]:
        """Returns the valid entity types that can be on the right side of a join."""
        return FrozenOrderedSet(
            join_type.right_entity_type for join_type in SemanticModelJoinLookup._VALID_ENTITY_JOINS
        )

    @cached_property
    def _model_id_to_has_validity_dimensions(self) -> Mapping[SemanticModelId, bool]:
        return {
            model_id: lookup.semantic_model.has_validity_dimensions
            for model_id, lookup in self._model_id_to_lookup.items()
        }

    def get_join_model_on_right_descriptors(
        self, left_model_id: SemanticModelId
    ) -> Mapping[SemanticModelId, OrderedSet[JoinModelOnRightDescriptor]]:
        """Return descriptors for semantic models that can be joined to `left_model_id`.

        The returned dict maps the ID of the semantic model on the right to the entities in the corresponding model
        that can be used as a join key.
        """
        right_model_id_to_join_descriptors: dict[
            SemanticModelId, MutableOrderedSet[JoinModelOnRightDescriptor]
        ] = defaultdict(MutableOrderedSet)
        left_model_lookup = self._model_id_to_lookup[left_model_id]
        left_model = left_model_lookup.semantic_model
        left_model_has_validity_dimensions = self._model_id_to_has_validity_dimensions[left_model_id]
        for entity in left_model_lookup.semantic_model.entities:
            left_entity_name = entity.name
            left_entity_type = entity.type
            other_model_lookups_with_same_entity_name = self._entity_name_to_model_ids[left_entity_name]
            for right_model_id in other_model_lookups_with_same_entity_name:
                right_model_lookup = self._model_id_to_lookup[right_model_id]
                right_model = right_model_lookup.semantic_model
                right_entity_type = right_model_lookup.entity_lookup.entity_name_to_type[left_entity_name]
                right_model_has_validity_dimension = self._model_id_to_has_validity_dimensions[right_model_id]

                join_type = EntityJoinType(
                    left_entity_type=left_entity_type,
                    right_entity_type=right_entity_type,
                )

                if join_type in SemanticModelJoinLookup._VALID_ENTITY_JOINS:
                    pass
                elif join_type in SemanticModelJoinLookup._INVALID_ENTITY_JOINS:
                    continue
                else:
                    raise ValueError(
                        LazyFormat(
                            "Unknown join type.",
                            join_type=join_type,
                            left_entity_name=left_entity_name,
                            left_model=left_model,
                            right_model=right_model,
                        )
                    )

                if left_model_has_validity_dimensions and right_model_has_validity_dimension:
                    # We cannot join two semantic models with validity dimensions due to concerns with unexpected fanout
                    # due to the key structure of these semantic models. Applying multi-stage validity window filters can
                    # also lead to unexpected removal of interim join keys. Note this will need to be updated if we enable
                    # simple-metric inputs in such semantic models, since those will need to be converted to a different type of semantic model
                    # to support such computation.
                    continue

                if right_entity_type is EntityType.NATURAL and not right_model_has_validity_dimension:
                    # There is no way to refine this to a single row per key, so we cannot support this join
                    continue

                right_model_id_to_join_descriptors[right_model_id].add(
                    JoinModelOnRightDescriptor(
                        entity_name=left_entity_name,
                        join_type=join_type,
                    )
                )
        return right_model_id_to_join_descriptors



from __future__ import annotations

import logging
import time
from typing import Iterable, Type

from typing_extensions import override

from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import MutableSemanticGraph, SemanticGraph
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class PartialSemanticGraphBuilder:
    """Builds a partial semantic graph using a subset of generators."""

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        self._manifest_object_lookup = manifest_object_lookup
        self._verbose_debug_logs = True

    def build(
        self,
        subgraph_generators: Iterable[Type[SemanticSubgraphGenerator]],
    ) -> SemanticGraph:
        """Return a semantic graph created by combining the output of the given generators."""
        current_graph = MutableSemanticGraph.create()
        for generator in subgraph_generators:
            start_time = time.perf_counter()
            generator_instance = generator(self._manifest_object_lookup)

            start_node_count = len(current_graph.nodes)
            start_edge_count = len(current_graph.edges)

            generated_edges = generator_instance.generate_edges()
            generated_edge_count = len(generated_edges)

            current_graph.add_edges(generated_edges)

            added_node_count = len(current_graph.nodes) - start_node_count
            added_edge_count = len(current_graph.edges) - start_edge_count
            runtime = time.perf_counter() - start_time

            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Generated subgraph",
                        generator=generator.__name__,
                        runtime=f"{runtime:.2f}s",
                        added_node_count=added_node_count,
                        added_edge_count=added_edge_count,
                        generated_edge_count=generated_edge_count,
                    )
                )

        return current_graph
